
import numpy as np




def pos_real_roots(p,imtol=1e-6):
	"""
	Return the sorted positive real roots of a polynomial.
	p is an array of coefficients of the polynomial
		p[n] * x^0 + p[n-1] * x^1 + ... + p[0] * x^n
	(aka highest order first).
	"""
	## get roots of polynomial
	r = np.roots(p)
	## find real values to within tolerance
	mask = ( np.abs(r.imag) < imtol )
	r = r[mask]
	r = np.real(r)
	## get positive values and sort
	r = np.sort( r[r>0] )
	## return
	return r












import numpy as np
import matplotlib.pyplot as plt
from pprint import pprint
import matplotlib.patches

def massplotrc():
	## tex
	plt.rcParams['text.usetex'] = True
	plt.rcParams['font.family'] = 'serif'
	plt.rcParams['font.serif'] = []
	plt.rcParams['text.latex.preamble'] = []
	## rc
	plt.rcParams['font.size'] = 7
	plt.rcParams['axes.linewidth'] = .4
	plt.rcParams['xtick.major.size'] = 2
	plt.rcParams['ytick.major.size'] = 2


def massplot(chainparams, params):
	"""
	"""
	## setup figure #################
	## figure size
	fig_width  = 2.5  ## inches
	fig_aspect = 1.68   ## height/width
	fig_height = fig_width * fig_aspect
	## define figure and axes
	plt.figure(99, figsize=(fig_width,fig_height))
	## axes
	h, hh = .08, .33
	ax1 = plt.axes([.08,1.-h-hh,.89,hh]) # [left,bottom,width,height]
	ax2 = plt.axes([.08,h      ,.89,hh]) # [left,bottom,width,height]
	ax = [ax1,ax2]
	## format axes
	for axx in [ax1,ax2]:
		plt.sca(axx)
		## axes
		plt.xticks([0,1])
		plt.yticks([0,1])
		plt.ylabel(r"$m/M$", labelpad=-4)
		plt.xlim(-.3,1.3)
		plt.ylim(-.1,1.1)
		plt.grid(1, alpha=.15, zorder=0)
	## labels
	ax1.set_xlabel(r"$v/\tau_{acc}$", labelpad=-6)
	ax2.set_xlabel(r"$u/\tau_{ev}$", labelpad=-6)
	## labels
	ax1.set_title("$(d)$ Formation Dynamics")
	ax2.set_title("$(e)$ Evaporation Dynamics")
	################################

	"""
	Parameters.
	"""

	## ideal total mass
	M = 0.5 * params['Rmax']
	
	## ideal lifetimes
	Tacc  = params['Tacc']
	Tevap = params['Tevap']

	## horizon radii
	Rh = chainparams['Rh']

	## initialize masses
	m = chainparams['m']

	## icrit=nacc and mcrit
	Nacc = params['Naccrete']
	icrit = np.argmax(m)
	mcrit = m[icrit]

	## junction locations
	r0, u0, v0 = chainparams['ps_r0'], chainparams['ps_u0'], chainparams['ps_v0']
	r1, u1, v1 = chainparams['fs_r0'], chainparams['fs_u0'], chainparams['fs_v0']

	## inf
	npoints = 5001
	inf = 100.

	"""
	Accretion.
	v is split into Nacc+1 intervals, each of length dv
	    vv = past starting point of each interval
	    mm = mass in each interval
	"""

	## init
	vacc  = 1.*np.zeros(Nacc+1)
	macc  = 1.*np.zeros(Nacc+1)

	## fill vacc values
	vacc[0] = -1.*inf
	dvacc = (v1-v0)[1:Nacc]
	vacc[1:] = np.cumsum(np.concatenate([np.array([0.]),dvacc]))

	## fill macc values
	macc = 1.*m[:Nacc+1]

	## reference vacc np.arrays
	vv = Tacc * np.linspace(-1, 2, npoints)

	## mass step function
	mmacc = 0.*vv
	for i in range(Nacc+1):
		mmacc[vv>vacc[i]] = macc[i]

	## ideal mass function
	x = 1.*vv/Tacc
	macc_ideal = 0.*vv
	macc_ideal[x>0.] = M*(0.5 + 0.5*vv[x>0.]/Tacc)
	macc_ideal[x>1.] = M + 0.*vv[x>1.]

	"""
	Evap.
	u is split into Nevap+1 intervals, each of length du
	    uu = past ending point of each interval
	    mm = mass in each interval
	"""

	## Nevap
	Nevap = len(m[icrit:])-1

	## init
	uevap  = 1.*np.zeros(Nevap+1)
	mevap  = 1.*np.zeros(Nevap+1)

	## fill uevap values
	uevap[0] = -1.*inf
	duevap = (u1-u0)[-Nevap:-1]
	uevap[1:] = np.cumsum(np.concatenate([ np.array([0.]), duevap]))

	## dvevap
	dvevap = (v1-v0)[-Nevap:-1]

	## fill mevap values
	mevap = 1.*m[-Nevap-1:]


	## reference uevap np.arrays
	uu = Tevap * np.linspace(-1, 2, npoints)

	## mass step function
	mmevap = 0.*uu
	for i in range(Nevap+1):
		mmevap[uu>uevap[i]] = mevap[i]

	## ideal mass function
	y = 1.*uu/Tevap
	mevap_ideal = 0.*uu
	mevap_ideal[y<1.] = M * (1. - (uu[y<1.]/Tevap))**(1./3.)
	mevap_ideal[y<0.] = 1.*M + 0.*uu[y<0.]


	"""
	Checks.
	"""


	## more checks
	print("r1-Rh")
	print(repr(r1-Rh))
	print("Rh/2m")
	print(repr(Rh/2*m))
	print("mcrit/M")
	print(repr(mcrit/M))

	## print checks
	print("m evap")
	print(repr(mevap))
	print("du evap")
	print(repr(duevap))
	print("dv evap")
	print(repr(dvevap))




	"""
	Plot.
	"""



	## style
	fill_sty = dict(alpha=.15, edgecolor='none', zorder=100)
	border_sty = dict(alpha=.3, lw=1, zorder=150)
	ideal_sty = dict(alpha=1., lw=1.5, ls='--')

	## ax1
	plt.sca(ax1)

	## plot accretion
	plt.fill_between(vv/Tacc, 0., mmacc/M, color='b', **fill_sty)
	plt.plot(vv/Tacc, mmacc/M, color='b', **border_sty)
	plt.plot(vv/Tacc, macc_ideal/M, color='b', **ideal_sty)

	## legend
	blue1, blue2 = (0.,0.,1.,.15), (0.,0.,1.,.4)
	a1, = plt.plot([], [], label=r"$m_v\!(v)$ (ideal)", color='b', linestyle="dashed", dashes=(3,3), lw=1.5)
	a2 = matplotlib.patches.Patch(label=r"$m_v\!(v)$ (numerical)", color=blue1, ec=blue2, lw=.7)

	## make legend
	leg1 = plt.legend(handles=[a1,a2], loc='lower right', numpoints=10, fontsize=6)
	leg1.zorder = 900
	leg1.get_frame().set_linewidth(.4)

	## ax2
	plt.sca(ax2)

	## plot evaportion
	plt.fill_between(uu/Tevap, 0., mmevap/M, color='r', **fill_sty)
	plt.plot(uu/Tevap, mmevap/M, color='r', **border_sty)
	plt.plot(uu/Tevap, mevap_ideal/M, color='r', **ideal_sty)

	## legend
	red1, red2 = (1.,0.,0.,.15), (1.,0.,0.,.4)
	a1, = plt.plot([], [], label=r"$m_u\!(u)$ (ideal)", color='r', linestyle="dashed", dashes=(3,3), lw=1.5)
	a2 = matplotlib.patches.Patch(label=r"$m_u\!(u)$ (numerical)", color=red1, ec=red2, lw=.7)

	## make legend
	leg1 = plt.legend(handles=[a1,a2], loc='lower left', numpoints=10, fontsize=6)
	leg1.zorder = 900
	leg1.get_frame().set_linewidth(.4)


	## save
	if False:
		plt.savefig("temp.png", dpi=800)
		plt.show()


	## show
	if False:
		plt.show()



####################### fake input #######################
def fakeinputs():
	"""
	"""
	## params
	params={'Tevap': 2.0,
		 'Tacc': 0.4,
		 'Naccrete': 5,
		 'Rmax': 0.5,
		 'Rmin': 0.2,
		 'dv_evap': 0.5,
		 'fparams0': {},
		 'fparams1': {'l': 0.01},
		 'l': 0.1,
		 'ueta': 0.0,
		 'uoff': 0.0,
		 'veta': 1.0,
		 'voff': -0.4}

	## chainparams
	chainparams= {'Rh': np.array([ 0.        ,  0.23932156,  0.29934074,  0.3593317 ,  0.41930663,
			        0.47927155,  0.45431214,  0.42702186,  0.39668451,  0.3622525 ,
			        0.32190157,  0.27169069,  0.19949748,  0.        ]),
			 'fs_r0': np.array([ 7.6411929 ,  7.56156025,  7.45288371,  7.35467021,  7.26557663,
			        0.57927155,  0.55431214,  0.52702186,  0.49668451,  0.4622525 ,
			        0.42190157,  0.37169069,  0.29949748,  7.68776392]),
			 'fs_u0': np.array([-15.6823858 , -17.00840717, -16.95371909, -16.89077914,
			       -16.82124943,   1.09869074,   1.52189764,   1.94109687,
			         2.35540545,   2.76383322,   3.16496844,   3.5563797 ,
			         3.93435595, -11.37552785]),
			 'fs_v0': np.array([-0.4 , -0.25, -0.15, -0.05,  0.05,  0.75,  1.25,  1.75,  2.25,
			        2.75,  3.25,  3.75,  4.25,  4.  ]),
			 'm': np.array([ 0.        ,  0.11987007,  0.14983759,  0.17980511,  0.20977263,
			        0.23974014,  0.22726618,  0.21362808,  0.19846838,  0.18126438,
			        0.16110626,  0.13602963,  0.1       ,  0.        ]),
			 'ps_r0': np.array([ 7.6411929 ,  7.6411929 ,  7.56156025,  7.45288371,  7.35467021,
			        7.26557663,  0.57927155,  0.55431214,  0.52702186,  0.49668451,
			        0.4622525 ,  0.42190157,  0.37169069,  0.29949748]),
			 'ps_u0': np.array([-15.6823858 , -17.27285923, -17.28010953, -17.19723407,
			       -17.11028635, -16.81971891,   0.76917269,   1.17997904,
			         1.58456074,   1.98070854,   2.36565531,   2.73424739,
			         3.07213381,   3.40100505]),
			 'ps_v0': np.array([-0.4 , -0.35, -0.25, -0.15, -0.05,  0.25,  0.75,  1.25,  1.75,
			        2.25,  2.75,  3.25,  3.75,  4.  ])}

	## return
	return chainparams, params
#######################################################################################


###########################################

if __name__=='__main__':
	## fake
	fake = True
	## go
	if fake:
		cp, p = fakeinputs()
		massplot(cp, p)
	## demo
	else:
		import xhorizon as xh
		xh.evap.demo()


###########################################




"""
This module contains functions for numerically evaluating the global tortoise function F(r),
and its inverse F^{-1}_{j}(rstar).

After the main block of code, there are supporting utilities and tests.
"""


import numpy as np
import scipy.integrate as spint



class F_by_interp:

	"""
	Our numerical implementation of the global tortoise function is always defined by linear interpolation
	of densely spaced reference arrays. This method is used even when an analytically defined alternative is
	possible (note that analytical alternatives are rarely available for Finv). The interpolation method is 
	generally valid, and allows F and Finv to be implemented symmetrically.

	The F_by_interp class is necessary in order to avoid accidentally overwriting the objects containing the
	reference arrays. Since a new F_by_interp instance is created when the class is called, each function
	F (and Finv) will have its own copy of the reference arrays.

	The class methods make_F() and make_Finv() generate functions F(r) and Finv(j, rstar) respectively:
		The inputs r and rstar should be arrays.
		The input j is the interval index label. It may be either a scalar (float or int), or an array of equal length to rstar.

	In order to create the reference arrays, F_by_interp calls create_reference, which in turn calls other subroutines.
	The call hierarchy is:
		F_by_interp < create_reference < F_by_integration < F_refpoints.

	The heavy lifting is done in F_by_integration, which implements the integral definition of the tortoise function.
	The function F_refpoints supports this task by preliminarily determining the reference values F(r_j+eps) in each interval.
	"""

	def __init__(self, f, rj, eps, npoints=500):

		## store input data
		self.f = f
		self.rj = rj
		self.eps = eps
		self.npoints = npoints

		## create reference r and rstar arrays
		ref = create_reference(f, rj, eps, npoints)

		## store reference arrays
		self.r_ref = ref[0]
		self.rstar_ref = ref[1]

		## initialize empty piecewise subarrays (filled by make_Finv)
		self.r_ref_j = []
		self.rstar_ref_j = []

	def make_F(self):

		"""
		Using the reference arrays, returns a function object F(r).
		Values outside the valid range return nan.

		Input r to F(r) should be an array.
		"""

		## define function by linear interpolation of reference array
		def F(r):
			rstar = np.interp(r, self.r_ref, self.rstar_ref, left=np.nan, right=np.nan)
			return rstar

		## return a function object
		return F

	def make_Finv(self):

		"""
		Using the reference arrays, returns a function object Finv(j, rstar).
		Values outside the valid range return nan.

		Inputs rstar and j to Finv(j,r) should be equal length arrays.

		If input j is a scalar (float or int), it is converted to an array of correct length with uniform value.
		"""

		## determine intervals and initialize reference subarrays
		rj = self.rj
		intervals = range(len(rj)-1)
		r_ref = self.r_ref
		rstar_ref = self.rstar_ref
		r_ref_j = []
		rstar_ref_j = []

		## make reference subarrays for each interval
		for k in intervals:
			mask = np.logical_and( r_ref>rj[k] , r_ref<rj[k+1] )
			r_ref_j += [r_ref[mask]]
			rstar_ref_j += [rstar_ref[mask]]

		## sort reference subarrays for each interval such that rstar is increasing
		for k in intervals:
			idx = np.argsort( rstar_ref_j[k] )
			rstar_ref_j[k] = rstar_ref_j[k][idx]
			r_ref_j[k] = r_ref_j[k][idx]

		## save reference subarrays
		self.r_ref_j = r_ref_j
		self.rstar_ref_j = rstar_ref_j

		## define function by linearly interpolating reference subarrays
		def Finv(j, rstar):

			## convert j to an array of integers
			if np.isscalar(j):
				j = j + np.zeros_like(rstar)
			j = j.astype(int)

			## initialize output as nan
			r = np.nan * rstar

			## cycle through intervals, interpolating relevant points in each one
			for k in intervals:
				mask = (j==k)
				r[mask] = np.interp(rstar[mask], self.rstar_ref_j[k], self.r_ref_j[k], left=np.nan, right=np.nan)

			## return
			return r

		## return a function object
		return Finv



def create_reference(f, rj, eps, npoints=500):
	"""
	Create a well-spaced array of r values and calculate the corresponding values rstar = F(r).
	Return both arrays to use as interpolation reference for forward and inverse tortoise functions.
	"""
	r = rspace1(rj, eps, npoints)
	rstar = F_by_integration(f, rj, eps, r)
	## return
	return r, rstar



def F_by_integration(f, rj, eps, r):
	"""
	Given the metric function f(r) and the values of r_j and epsilon,
	returns an array rstar=F(r) from an array r.
	The values are determined by performing the defining definite integral.
	"""
	## initialize output to nan
	rstar = np.nan * r
	## get reference values
	Fj = F_refpoints(f, rj, eps)
	## determine appropriate interval for each point in r
	jj = which_interval(r, rj)
	## define integrand for calculation and loop over each point in r
	integrand = lambda r: f(r)**(-1)
	for m in range(len(r)):
		j = jj[m]
		a = rj[j] + eps
		b = r[m]
		rstar[m] = Fj[j] + integrate(integrand, a, b)
	## return
	return rstar



def F_refpoints(f, rj, eps):
	"""
	Given the metric function f(r) and the values of r_j and epsilon,
	returns an array of of the values F(r_j+eps) == F_j.
	Special Values:
		F(eps) = F(r_0 + eps) = 0
		F(inf + eps) = F(r_{N+1} + eps) = nan
		F(r_i + eps) = F(r_i - eps)
	Recursive Formula:
		F_0 = 0
		F_k = F_{k-1} + int_{r_{k-1}+eps}^{r_{k}-eps}
	"""
	## initialize output array to nan
	Fj = np.nan * rj
	## first value is necessarily zero
	Fj[0] = 0.
	## calculate remaining values by integrating
	integrand = lambda r: f(r)**(-1)
	for k in range(1, len(rj)-1 ):
		a = rj[k-1] + eps
		b = rj[k  ] - eps
		Fj[k] = Fj[k-1] + integrate(integrand, a, b)
	## return
	return Fj











"""
Utilities.
"""

def integrate(func,a,b):
	"""
	Wrapper for definite integration using scipy.integrate.quad().
	"""
	## calculate integral and error
	val, abserr = spint.quad(func,a,b, limit=100)
	## get relative error
	relerr = np.nan
	if val != 0:
		relerr = np.abs(abserr/val)
	## warning parameters
	maxrel, maxabs = 1e-4, 1e-6
	## execute warnings
	if (val!=0. and relerr > maxrel):
		print("\nIntegration Warning: Relative error exceeds %e"%(maxrel))
		print("%s from %s to %s"%(func,a,b))
		print("val %s, abserr %e, relerr %e"%(val,abserr,relerr))
	if (np.abs(val) < 1. and abserr > maxabs):
		print("\nIntegration Warning: Absolute error exceeds %e"%(maxabs))
		print("%s from %s to %s"%(func,a,b))
		print("val %s, abserr %e, relerr %e"%(val,abserr,relerr))
	## return
	return val


def which_interval(r,rj,asint=True):
	"""
	Given the radii r_j, determine which interval I_j a radius belongs to. Return nan if error.
	Inputs are the array rj and an arbitrary length array r.
	Output is an array j of equal length as r giving the corresponding interval
	for each radius in array r.
	"""
	j = np.nan * r
	for jj in range(len(rj)-1):
		mask = np.logical_and(r>rj[jj], r<rj[jj+1])
		j[mask] = jj
	if np.any(np.isnan(j)):
		print("\nWARNING: Function which_interval() encountered invalid values. Returning nan.\n")
	if asint==True:
		j = j.astype(int)
	return j


def rspace0(r0, r1, eps, npoints=500):
	"""
	Provide a dense array of r values for interpolating F and Finv.
	Must have extra values near exponential tails to smoothly interpolate there.
	"""
	## interval
	a = r0 + eps
	b = r1 - eps
	## linear spacing
	rr0 = np.linspace(a, b, 2*npoints)
	## exponential near tails
	B = np.log(eps**2)
	rr1 = a + 0.5 * (b-a) * np.exp( np.linspace(0, B, npoints) )
	rr2 = b - 0.5 * (b-a) * np.exp( np.linspace(0, B, npoints) )
	## concatenate and sort
	rr = np.sort(np.concatenate( [rr0,rr1,rr2] ))
	## return
	return rr


def rspace1(rj, eps, npoints=500):
	"""
	Like rspace0(), but gives values over all intervals I_j simultaneously.
	"""
	rx = []
	for j in range(len(rj)-1):
		rx += [rspace0(rj[j], rj[j+1], eps, npoints)]
	rvals = np.sort(np.concatenate(rx))
	## return
	return rvals


def first_trapped_j(f,rj):
	r0 = 0.5 * ( rj[0] + rj[1] )
	f0 = f(r0)
	out = np.nan
	if f0>0.:
		out = 1
	if f0<0.:
		out = 0
	return out









"""
Tests. 
Tests run if __name__ == "__main__".
"""

import matplotlib.pyplot as plt

def test1():
	"""
	Tests whether the function scipy.quad(h,a,b) obeys int_a^b = - int_b^a, even when endpoints
	approach singular points of the integral.
	This would be unlikely for numerical integration of an ODE, but for this area-based 
	quadrature it appears to be true.
	In fact, it appears to be true even when the integral method provides a warning.
	Since this is true, integrals of the form int_{rj+eps}^b are admissible, and using a
	less extreme integration point in the middle of the interval will not increase accuracy.
	"""
	######### input ########
	f = lambda r: 1. - 1. / r
	h = lambda r: f(r)**(-1)
	a = 1.5
	b = 1. + 1e-11
	########################
	print("\nTEST 1")
	forward      =  integrate(h,a,b)
	neg_backward = -integrate(h,b,a)
	abs_diff = forward - neg_backward
	sumtot = forward + neg_backward
	print(" forward  = % .18e"%forward)
	print("-backward = % .18e"%neg_backward)
	print(" abs diff = % .18e"%abs_diff)
	if sumtot != 0.:
		rel_diff = 2. * abs_diff / sumtot
		print(" rel diff = % .18e"%rel_diff)
	## end
	print("END TEST 1\n")




def test2():
	"""
	Demonstrate the functionality of F_refpoints.
	"""
	######### input ########
	f = lambda r: ( 1. - 1. / r )
	rj = np.array([0.,1.,25.])
	eps = 1e-11
	########################
	print("\nTEST 2")
	Fj = F_refpoints(f,rj,eps)
	print("Fj = ", Fj)
	## end
	print("END TEST 2\n")


def test3():
	"""
	Test of which_interval().
	"""
	######### input ########
	r = np.linspace(1e-12, 15, 20)
	rj = np.array([0.,1.,2.,8.,25.])
	########################
	print("\nTEST 3")
	j = which_interval(r, rj)
	print("j = ", j)
	## end
	print("END TEST 3\n")


def test4():
	"""
	Demonstrate the functionality of F_by_integration.
	"""
	######### input ########
	f = lambda r:  1. - 1. / r 
	rj = np.array([0.,1.,25.])
	eps = 1e-9
	r = np.linspace(1e-9,7,1000)
	########################
	print("\nTEST 4")
	## numerically calculate values
	rstar = F_by_integration(f,rj,eps,r)
	## know analytical solution
	rstar2 = r + np.log(np.abs(r-1.))
	## plot
	plt.plot(r, rstar,  'bx',  zorder=10)
	plt.plot(r, rstar2, 'r', lw=1.3, zorder=20)
	plt.grid()
	plt.show()
	## end
	print("END TEST 4\n")



def test5():
	"""
	Demonstrate the functionality of rspace0().
	"""
	######### input ########
	r0 = 0.
	r1 = 1.
	eps = 1e-9
	npoints = 500
	########################
	print("\nTEST 5")
	rvals = rspace0(r0,r1,eps,npoints)
	print("rvals = ", rvals)
	plt.plot(rvals,'kx')
	plt.show()
	## end
	print("END TEST 5\n")


def test6():
	"""
	Demonstrate the functionality of rspace1().
	"""
	######### input ########
	rj = np.array([0.,1.,10.,25.])
	eps = 1e-9
	npoints = 500
	########################
	print( "\nTEST 6")
	rvals = rspace1(rj,eps,npoints)
	print( "rvals = ", rvals)
	plt.plot(rvals,'kx')
	plt.show()
	## end
	print( "END TEST 6\n")


def test7():
	"""
	Demonstrate the functionality of create_reference().
	"""
	######### input ########
	f = lambda r: 1. -  1./r
	rj = np.array([0.,1.,25.])
	eps = 1e-9
	npoints=500
	########################
	print("\nTEST 7")
	## go
	r, rstar = create_reference(f, rj, eps, npoints)
	## plot
	plt.plot(r,rstar,'kx')
	plt.grid()
	plt.show()
	## end
	print("END TEST 7\n")


def test8():
	"""
	Demonstrate the functionality of F_by_interp.make_F().
	"""
	######### input ########
	f = lambda r: 1. -  1./r
	rj = np.array([0.,1.,25.])
	eps = 1e-9
	npoints=500
	r = np.concatenate([rspace1(rj,eps,npoints), np.linspace(50,-50,1000)])
	########################
	print("\nTEST 8")
	## build F_by_interp
	print("only this part is slow....")
	tort_by_interp = F_by_interp(f, rj, eps, npoints)
	print("see?")
	## make F
	F = tort_by_interp.make_F()
	## plot
	plt.plot(r,F(r),'kx')
	plt.grid()
	plt.show()
	## end
	print("END TEST 8\n")


def test9():
	"""
	Demonstrate the functionality of F_by_interp.make_Finv().
	"""
	######### input ########
	f = lambda r: 1. -  1./r
	rj = np.array([0.,1.,2.,25.])
	eps = 1e-9
	npoints=100
	rstar = np.linspace(-10,50,5000)
	jvals = [0,1,2]
	########################
	print("\nTEST 9")
	## build F_by_interp
	print("only this part is slow....")
	tort_by_interp = F_by_interp(f, rj, eps, npoints)
	print("see?")
	## make F
	Finv = tort_by_interp.make_Finv()
	## plot
	for j in jvals:
		plt.plot(Finv(j,rstar), rstar,'x')
	plt.grid()
	plt.show()
	## end
	print("END TEST 9\n")




## run tests if __name__="__main__"

if __name__=="__main__":
	print("RUN TESTS")
	test1()
	test2()
	test3()
	test4()
	test5()
	test6()
	test7()
	test8()
	test9()





from xhorizon.shell_junction import helpers
from xhorizon.shell_junction import interpolators
from xhorizon.shell_junction import reg_corner_masks
from xhorizon.shell_junction import active_slice_class
from xhorizon.shell_junction import passive_slice_class
from xhorizon.shell_junction import tests



from xhorizon.shell_junction.active_slice_class import active_slice as aslice
from xhorizon.shell_junction.passive_slice_class import passive_slice as pslice

from xhorizon.shell_junction.helpers import set_ruv0, schwarz_dtr_masses, slicecheck




"""
This module contains subroutines for filling parts of the diagram with a background color.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

from xhorizon.diagram_tools import curve_class as cc
from xhorizon.diagram_tools.curvemakers import rstarlines_special_2, block_boundary, block_boundary_2, rlines, rstarlines




def fill_between_curves_uv(blk, crvlist, tr=False, reverse=True, sty={}, eps=1e-24):
	"""
	Fill in a patch using a list of curves to provide the boundary of polygon.
	The curves are read from their uv coords.
	Every other input curve is reversed by default.
		block = block in which to fill curves
		crvlist = list of two curves with uv coords
		tr = set to true if tr coords are primary
		reverse = reverse point order of every other curve
	Curves are clipped to given rectangular size using snap_crvlist_to_bounds.
	"""
	## default style
	style = dict(fc='c', ec='none', lw=0.25, zorder=100)
	## get uvbounds from block
	uv=dict(umin=-np.inf, umax=np.inf, vmin=-np.inf, vmax=np.inf)
	uv.update(blk.uvbounds)
	umin, umax = uv['umin'], uv['umax']
	vmin, vmax = uv['vmin'], uv['vmax']
	## make sure limits are float
	umin, umax, vmin, vmax = float(umin), float(umax), float(vmin), float(vmax)
	## reverse odd index input curves
	if reverse==True:
		for i in range(len(crvlist)):
			if (i % 2) == 1:
				crvlist[i] = reverse_curve(crvlist[i])
	## start from tr values if requested
	if tr==True:
		crvlist = blk.update_curves_from_tr(crvlist)
	## snap uv values to within desired rectangle
	crvlist = snap_crvlist_to_bounds(crvlist, umin=umin, umax=umax, vmin=vmin, vmax=vmax, eps=eps)
	## update crvlist from new uv coords and apply masks
	crvlist = blk.apply_masks( blk.update_curves_from_uv( crvlist ) )
	## concatenate UV coords to create boundary
	U = np.concatenate([ crv.UV[0] for crv in crvlist ])
	V = np.concatenate([ crv.UV[1] for crv in crvlist ])
	## create xy array
	x, y = V-U, V+U
	xy = np.dstack([x,y])[0]
	## plot polygon
	style.update(sty)
	p = mpl.patches.Polygon(xy, **style)
	plt.gca().add_patch(p)



def fill_block(blk, sty={}):
	"""
	Fill in an entire block.
	"""
	## get uvbounds from block
	uv=dict(umin=-np.inf, umax=np.inf, vmin=-np.inf, vmax=np.inf)
	uv.update(blk.uvbounds)
	umin, umax = uv['umin'], uv['umax']
	vmin, vmax = uv['vmin'], uv['vmax']
	## default style
	style = dict(fc='k', ec='none', zorder=10)
	style.update(sty)
	## generate block boundaries
	crvlist = block_boundary_2(blk)
	## fill between boundaries
	fill_between_curves_uv(blk, crvlist, sty=sty)



def fill_between_r(blk, rvals=np.array([0.,0.]), sty={}, inf=100., npoints=1000):
	"""
	"""
	## default to invalid params
	valid = False
	## check that there is something to fill
	eps = blk.master.metfunc.Fparams['eps']
	rsnap = snap_to_bounds(rvals, ymin=blk.rj[0]+1.1*eps, ymax=blk.rj[1]-1.1*eps)
	if np.abs(rsnap[1]-rsnap[0]) > 5.*eps:
		valid = True
	## get block boundary curves
	boundary_curves = block_boundary_2(blk, inf=inf, npoints=npoints)
	## get params and initialize boundary rstar values
	boundary_rstar = np.nan * rvals
	c = float(blk.master.rparams['c'])
	## determine block boundary rstar values in order of increasing radius
	for i in range(2):
		u,v = boundary_curves[i].uv
		boundary_rstar[i] = 0.5*(v[0]-u[0])+c
	## for each rval either use F(r) or boundary value as rstar
	rvals = np.sort(rvals)
	rstar_vals = np.nan * rvals
	for i in range(len(rvals)):
		rval = rvals[i]
		## check if radius value is finite
		if not np.isfinite(rval):
			valid = False
		## if radius too small use left boundary
		elif rval <= blk.rj[0]+1.1*eps:
			rstar_vals[i] = boundary_rstar[0]
		## if radius too big use right boundary
		elif rval >= blk.rj[1]-1.1*eps:
			rstar_vals[i] = boundary_rstar[1]
		## if radius is valid use F(r)
		else:
			rstar_vals[i] = blk.master.metfunc.F(rval)
	## check that new values are finite
	if not np.all(np.isfinite(rstar_vals)):
		valid = False
	## go if valid
	if valid==True:
		## can be buggy if inf or npoints are too low
		#crvlist = rstarlines(rstar_vals, c=c, inf=2.*inf, npoints=npoints, sty=dict(marker='x'))
		crvlist = rstarlines_special_2(rstar_vals, blk.uvbounds, c=1.*c, inf=inf, npoints=npoints, eps=1e-15)
		fill_between_curves_uv(blk, crvlist, sty=sty, eps=eps)






"""
Useful utilities for filling.
"""



def reverse_curve(crv):
	"""
	Reverse the order of points in a curve.
	"""
	## reverse each 1d subarray
	crv.r    =    crv.r[::-1]
	crv.tr   =   crv.tr[:,::-1]
	crv.uv   =   crv.uv[:,::-1]
	crv.uvdl = crv.uvdl[:,::-1]
	crv.UV   =   crv.UV[:,::-1]
	## return
	return crv




def snap_to_bounds(y, ymin=-np.inf, ymax=np.inf):
	"""
	Limit an array y to the range ymin <= y <= ymax by replacing values 
	outside the range with the cap values.
	"""
	## copy y into z
	z = 1.*y
	## snap z to bounds
	z[y<ymin] = ymin + 0. * z[y<ymin] 
	z[y>ymax] = ymax + 0. * z[y>ymax] 
	## return
	return z


def snap_crvlist_to_bounds(crvlist, umin=-np.inf, umax=np.inf, vmin=-np.inf, vmax=np.inf, eps=1e-15):
	"""
	Apply snap_to_bounds to uv coordinates of a curve.
	Displaces curve into a rectangle of specified size.
	Move inward by eps to avoid edge cases.
	"""
	## define new curve list
	crvlist2 = []
	## alter min and max values
	umin, umax = umin+eps, umax-eps
	vmin, vmax = vmin+eps, vmax-eps
	## snap to bounds
	for crv in crvlist:
		crv.uv[0] = snap_to_bounds(crv.uv[0], ymin=umin, ymax=umax)
		crv.uv[1] = snap_to_bounds(crv.uv[1], ymin=vmin, ymax=vmax)
		crvlist2 += [crv]
	## return
	return crvlist2








"""
Test routines. Will run if __name__=='__main__'.
"""

def test1():
	"""
	Test matplotlib polygon patch functionality.
	"""
	##
	x1 = np.linspace(-3,3,10000)
	y1 = np.sin(2.*np.pi*x1)
	##
	x = np.concatenate([x1,y1])
	y = np.concatenate([y1,x1])
	xy = np.dstack([x,y])[0]
	##
	plt.figure()
	plt.xlim(-4,4)
	plt.ylim(-4,4)
	plt.gca().set_aspect('equal')
	##
	sty = dict(fc='c', ec='none', alpha=.5)
	p = mpl.patches.Polygon(xy, **sty)
	plt.gca().add_patch(p)
	##
	plt.show()


def test2():
	"""
	Test of snap_to_bounds.
	"""
	## define arrays to clip
	x = np.linspace(0,10,1000)
	y = np.sin(x)
	## clip
	y1 = snap_to_bounds(y)
	y2 = snap_to_bounds(y, ymin=-.8, ymax=.7)
	## plot
	plt.plot(x,y1,'b-',lw=15)
	plt.plot(x,y1,'r-',lw=10)
	plt.plot(x,y2,'y-',lw=5)
	## show
	plt.show()



if __name__=='__main__':
	print("TESTS")
	#test1()
	#test2()












import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from xhorizon.evap.draw import *
import matplotlib.colors


## legend
def legend():

	## setup legend #################
	## tex
	plt.rcParams['text.usetex'] = True
	plt.rcParams['font.family'] = 'serif'
	plt.rcParams['font.serif'] = []
	plt.rcParams['text.latex.preamble'] = []
	## rc
	fs = 5
	plt.rcParams['font.size'] = fs
	plt.rcParams['axes.linewidth'] = .4
	## figure size
	fig_height = 1.68 * 2.5  ## inches
	fig_width = 1.54  ## inches
	## axis loc
	## define figure and axes
	plt.figure(1, figsize=(fig_width, fig_height))
	## axes loc
	x = .31
	ax1 = plt.axes([.025,x+.047,.95,.43 ]) ## left bottom width height
	ax2 = plt.axes([.075,x,     .85,.037]) ## left bottom width height
	## format axes
	for axx in [ax1,ax2]:
		## set axis
		plt.sca(axx)
		## lims
		plt.xlim(0,1)
		plt.ylim(0,1)
		## ticks
		plt.xticks([])
		plt.yticks([])
		## off
		#plt.axis('off')
	################################

	if True:
		## legend
		plt.sca(ax1)

		## init
		labels = []
		hand   = []


		## r=0
		labels += [r"$r=0$ (no singularity)"]
		hand += [mpl.lines.Line2D([], [], **rline_zero_sty)]
		## r=0 singularity
		labels += [r"$r=0$ (singularity)"]
		rline_zero_sty.update(singularity_sty)
		hand += [mpl.lines.Line2D([], [], **rline_zero_sty)]

		## blank
		labels += [r""]
		hand += [mpl.lines.Line2D([], [], c='w')]

		## horizons
		labels += [r"horizons ($f(r)=0$)"]
		hand += [mpl.lines.Line2D([], [], **rline_hor_sty)]
		## trapped region
		labels += [r"trapped spheres region"]
		hand += [mpl.patches.Patch(fc='w', ec='k', hatch='.', lw=.3)]

		## blank
		labels += [r""]
		hand += [mpl.lines.Line2D([], [], c='w')]

		## shells
		labels += [r"positive-mass shells ($\sigma>0$)"]
		acc_shells_sty.update(dict(dashes=(2.5,2.5), lw=.8))
		hand += [mpl.lines.Line2D([], [], c='0.5', **acc_shells_sty)]
		## shells
		labels += [r"negative-mass shells ($\sigma<0$)"]
		evap_shells_in_sty.update(dict())
		hand += [mpl.lines.Line2D([], [], c='0.5', **evap_shells_in_sty)]
		## shells
		labels += [r"(shell darkness $ \sim \Delta m /M$)"]
		hand += [mpl.lines.Line2D([], [], c='w')]

		## blank
		labels += [r""]
		hand += [mpl.lines.Line2D([], [], c='w')]

		## r = const
		labels += [r"$r=\textrm{constant} \;\; (dr = l/2)$"]
		alpha = rline_sty['alpha']
		col = np.array(mpl.colors.colorConverter.to_rgb('c'))
		col2 = 0.75*np.array(mpl.colors.colorConverter.to_rgb(plt.cm.Oranges(.1))) + .25
		col = alpha*col + (1.-alpha)*col2
		rline_sty.update(c=col, alpha=1.)
		hand += [mpl.lines.Line2D([], [], **rline_sty)]
		## r = const
		labels += [r"$r=\textrm{constant} \;\; (dr = 2M/2)$"]
		col = np.array(mpl.colors.colorConverter.to_rgb('m'))
		col2 = 0.75*np.array(mpl.colors.colorConverter.to_rgb(plt.cm.Oranges(.1))) + .25
		col = alpha*col + (1.-alpha)*col2
		rline_sty.update(c=col, alpha=1.)
		hand += [mpl.lines.Line2D([], [], **rline_sty)]

		## blank
		labels += [r""]
		hand += [mpl.lines.Line2D([], [], c='w')]

		## r = inf
		labels += [r"$r=\infty \;\; (\textrm{width} \! \propto \! (1 \! + \! 2m/M) )$"]
		rline_inf_sty2 = rline_inf_sty.copy()
		rline_inf_sty2.update(dict(lw=3, ls='dashed', dashes=(6,4)))
		hand += [(mpl.lines.Line2D([], [], **rline_inf_sty ),
		          mpl.lines.Line2D([], [], **rline_inf_sty2) )]

		## ticks
		labels += [r"$d\tau_{\infty}=\textrm{constant}$"]
		tsty = dict(markeredgecolor='0.5', ls='none', marker=(2,0,0), markersize=5)
		hand += [mpl.lines.Line2D([], [], **tsty)]



		## make legend
		legend = plt.legend(hand, labels, loc='center', borderpad=1, labelspacing=.75, fontsize=fs, numpoints=4, frameon=False)
		legend.get_frame().set_linewidth(.5)





	if True:
		## colorbar
		plt.sca(ax2)

		## params
		npoints = 200
		cmin, cmax, alpha = .1, .6, .75

		## custom colors
		cc = np.linspace(cmin, cmax, npoints)
		cols = alpha * plt.cm.Oranges(cc) + (1.-alpha)
		cols[:,-1] = 1. + 0.*cols[:,-1]

		## colormap norm etc
		cmap = matplotlib.colors.ListedColormap(cols)
		norm = mpl.colors.Normalize(vmin=0., vmax=1.)
		ticks = []
		orientation = 'horizontal'

		##
		plt.annotate(s=r"proper density ($\rho/\rho_0$)", xy=(.5,-.35), xycoords='axes fraction', ha='center', va='top', size=fs)
		plt.annotate(s=r"$0$", xy=(0.01,-.22), xycoords='axes fraction', ha='center', va='top', size=fs)
		plt.annotate(s=r"$1$", xy=(1.,-.22), xycoords='axes fraction', ha='center', va='top', size=fs)


		## colorbar
		cb1 = mpl.colorbar.ColorbarBase(ax2, cmap=cmap, norm=norm, ticks=ticks, orientation=orientation)



	## save
	if True:
		path = "temp-figs/demo"
		plt.savefig("%s_legend.png"%(path), dpi=800)



if __name__=="__main__":
	legend()




"""
This module contains helpers to draw with evap code.
"""

import numpy as np
import matplotlib.pyplot as plt
import copy
import matplotlib.hatch
import matplotlib as mpl
import matplotlib.colors

import xhorizon as xh
from xhorizon.shell_junction import interpolators as interp
from xhorizon.evap.helpers import irr


############## plot reglist ####################


## styles
rline_zero_sty      = dict(lw=1.5, c='0.48', zorder=11000)
singularity_sty     = dict(ls='dashed', dashes=(2,2))
rline_inf_sty       = dict(lw=1.5, c='0.48', zorder=10000)
rline_hor_sty       = dict(lw=0.4, c='k', zorder=9999)
rline_sty           = dict(ls='-', zorder=10, lw=.6, alpha=.5)
acc_shells_sty      = dict(lw=0.6, ls='dashed', dashes=(3,3), zorder=2000)
evap_shells_out_sty = acc_shells_sty
evap_shells_in_sty  = dict(lw=1., ls='dashed', dashes=(1,1.2), zorder=2000)
fill_horizons_sty   = dict(fc='none', ec='k', lw=0, hatch='c', zorder=9990)
fill_density_sty    = dict(zorder=100)
tick_sty            = dict(markersize=10, markeredgecolor='k', alpha=.3, zorder=100)


def drawreg(reglist, chainparams, fparams=dict()):
	"""
	Plot all regions in reglist.
	"""       
	l=fparams['l']
	R=fparams['R']

	## lines
	rline_zero(reglist, sty=rline_zero_sty, sty2=singularity_sty, npoints=1001)
	rline_inf_col(reglist, sty=rline_inf_sty, npoints=5001, inf=100., M=0.5*R)
	rline_hor(reglist, sty=rline_hor_sty)
	acc_shells(reglist, chainparams, sty=acc_shells_sty, inf=100., npoints=1001)
	evap_shells_out(reglist, chainparams, sty=evap_shells_out_sty, inf=100., npoints=1001)
	evap_shells_in(reglist, chainparams, sty=evap_shells_in_sty, inf=100., npoints=1001)
	make_rlines(reglist, chainparams, l=l, R=R, Tevap=5., sty=rline_sty)
	vticks(reglist, dv=0.5, sty=tick_sty)
	uticks(reglist, du=0.5, sty=tick_sty)
	tticks(reglist, dt=0.5, sty=tick_sty)
	## plot
	for reg in reglist:
		reg.rplot()
	## fill
	fill_horizons(reglist, sty=fill_horizons_sty)
	fill_density(reglist, sty=fill_density_sty)




#############################################################


## being actually used ############

def rline_zero(reglist, npoints=5001, inf=100., sty={}, sty2={}):
	"""
	Add boundary lines to regions in reglist.
	"""
	for reg in reglist:
		for b in reg.blocks:
			if b.j==0:
				## style
				style = dict(lw=0.9, c='0.5', zorder=11000)
				style.update(sty)
				## singularity style
				if b.sgnf<0.:
					style.update(sty2)
				## cutoff style
				if reg.metfunc.info['Type'] in ['Hayward - de Sitter with Cutoff']:
					lw = 1.5*(1.+ 2.)
					style.update(lw=lw)
				## curve
				cv = xh.cm.rstarlines_special_2([0.], b.uvbounds, c=1.*reg.rparams['c'], sty=style, inf=1.*inf, npoints=1.*npoints, eps=1e-12)
				b.add_curves_uv(cv)
				




def rline_inf_col(reglist, npoints=5001, inf=100., sty={}, M=.5):
	"""
	Add boundary lines to regions in reglist.
	"""
	for reg in reglist:
		for b in reg.blocks:
			## outermost blocks only
			if b.j==len(b.master.metfunc.rj)-2:
				## style
				style = dict(c='0.5', zorder=10000)
				style.update(sty)
				## rstar value
				rstar = b.master.metfunc.rstar_ref[-1]
				## mass
				m = 0.
				if 'R' in b.master.metfunc.fparams.keys():
					m = 0.5 * b.master.metfunc.fparams['R']
				## linewidth
				lw0 = 1.*style['lw']
				lw = lw0 * (1.+2.*m/M)
				style.update(dict(lw=lw))
				## curve
				style.update(solid_capstyle='projecting')
				b.add_curves_uv(xh.cm.block_boundary_2(b, sty=style, inf=2., npoints=npoints)[-1:])
				style.update(solid_capstyle='butt')
				b.add_curves_uv(xh.cm.block_boundary_2(b, sty=style, inf=inf, npoints=npoints)[-1:])


def rline_hor(reglist, npoints=5001, sty={}):
	"""
	Add horizon lines to regions in reglist.
	"""
	for reg in reglist:
		for b in reg.blocks:
			## trapped blocks only
			if b.sgnf<0.:
				## style
				style = dict(lw=0.2, c='k', zorder=9999)
				style.update(sty)
				## rstar value
				rstar = b.master.metfunc.rstar_ref[-1]
				## curve
				cvs = xh.cm.block_boundary_2(b, sty=style, inf=100., npoints=npoints)
				if b.j==0:
					cvs = cvs[-1:]
				## add
				b.add_curves_uv(cvs)

def shell_col(x):
	Cmax = .8
	return np.array([1.,1.,1.]) - 0.33*(1.+2.*x)*Cmax

def acc_shells(reglist, chainparams, sty={}, inf=50., npoints=5001):
	"""
	Add boundary lines to regions in reglist.
	"""
	Rh = chainparams['Rh']
	v0 = chainparams['fs_v0']-1e-15
	m = chainparams['m']
	M = np.max(m)
	for i in range(len(reglist[:-1])):
		reg = reglist[i]
		## accretion only
		if Rh[i+1]>Rh[i]:
			##
			for b in reg.blocks:
				## mass
				dm = m[i+1] - m[i]
				print("ACC SHELL COLOR dm/M=%s"%(dm/M))
				col = shell_col(dm/M)
				## style
				style = dict(lw=0.2, ls='dashed', dashes=(4,4), c=1.*col, zorder=2000)
				style.update(sty)
				## v value
				vv = v0[i:i+1]
				## curve
				b.add_curves_uv(xh.cm.uvlines(vv, uv='v', uvbounds=b.uvbounds, sty=style, c=0., inf=inf, npoints=npoints))


def evap_shells_out(reglist, chainparams, sty={}, inf=100., npoints=5001):
	"""
	Add boundary lines to regions in reglist.
	"""
	Rh = chainparams['Rh']
	u0 = chainparams['fs_u0']-1e-15
	m = chainparams['m']
	M = np.max(m)
	for i in range(len(reglist[:-1])):
		reg = reglist[i]
		## evap only
		if Rh[i+1]<Rh[i]:
			## outgoing
			for b in reg.blocks:
				if b.bparams['epsu']>0.:
					if not np.isfinite(b.uvbounds['vmax']):
						## mass
						dm = - (m[i+1] - m[i])
						print("EVAP OUT SHELL COLOR dm/M=%s"%(dm/M))
						col = shell_col(dm/M)
						## style
						style = dict(lw=0.2, ls='dashed', dashes=(4,4), c=1.*col, zorder=2000)
						style.update(sty)
						## u value
						uu = u0[i:i+1]
						## curve
						b.add_curves_uv(xh.cm.uvlines(uu, uv='u', uvbounds=b.uvbounds, sty=style, c=0., inf=inf, npoints=npoints))


def evap_shells_in(reglist, chainparams, sty={}, inf=5., npoints=5001):
	"""
	Add boundary lines to regions in reglist.
	"""
	Rh = chainparams['Rh']
	u0 = chainparams['fs_u0']
	v0 = chainparams['fs_v0']-1e-15
	r0 = chainparams['fs_r0']
	m = chainparams['m']
	M = np.max(m)
	for i in range(len(reglist[:-1])):
		reg = reglist[i]
		## evap only
		if Rh[i+1]<Rh[i]:
			## uv value
			uu = u0[i:i+1]
			vv = v0[i:i+1]
			rr = r0[i:i+1]
			## ingoing
			for b in reg.blocks:
				## mass
				dm = - (m[i+1] - m[i])
				print("EVAP IN SHELL COLOR dm/M=%s"%(dm/M))
				col = shell_col(dm/M)
				## style
				style = dict(lw=0.9, ls=':', c=1.*col, zorder=2000)
				style.update(sty)
				## inner blocks
				if b.rj[1]<rr:
					## values
					u = np.sort(np.concatenate([np.linspace(-1e-3,-1.,npoints//2),np.linspace(1e-3,1.,npoints//2),np.linspace(-inf,inf,npoints)]))
					v = vv+0.*u
					## make curve
					cv = xh.curve()
					cv.uv = np.array([u,v])
					cv.sty.update(style)
					## add
					b.add_curves_uv([cv])
				## outer block
				elif b.rj[0]<rr and not np.isfinite(b.uvbounds['umax']):
					## values
					u = np.linspace(uu,inf,npoints)
					v = vv+0.*u
					## make curve
					cv = xh.curve()
					cv.uv = np.array([u,v])
					cv.sty.update(style)
					## add
					b.add_curves_uv([cv])




def fill_horizons(reglist, sty={}):
	##
	for i in range(len(reglist)):
		reg = reglist[i]
		for b in reg.blocks:
			if b.sgnf < 0.:
				style = dict(fc='none', ec='k', lw=0, hatch='.', zorder=9990)
				style.update(sty)
				b.fill(sty=style)


def fill_density(reglist, sty={}):
	## regions
	for reg in reglist:
		## linespacing
		x = np.arange(0.,10.1,.5)
		## default to zero
		density = 0.*x
		## vac spacetimes
		vac = ['Schwarzschild', 'Minkowski', 'Anti de Sitter', 'de Sitter', 'Hayward - de Sitter with Cutoff']
		if reg.metfunc.info['Type'] in vac:
			for b in reg.blocks:
				b.fill(dstyle(0.))
		## hayward
		if reg.metfunc.info['Type'] in ['Hayward', 'Hayward - Anti de Sitter', 'Hayward - de Sitter']:
			## get params
			l = reg.metfunc.fparams['l']
			R = reg.metfunc.fparams['R']
			## rcore and r
			rcore = (R*l**2.)**(1./3.)
			r = rcore * x
			rmax = 100.
			r = np.concatenate([r, np.array([rmax])])
			## density 
			density = 1./(1.+x**3.)**2.   # rho/rho0 = 1/(1+x^3)^2
			## blocks
			for b in reg.blocks:
				## loop over radii
				for n in range(len(x)):
					## rr and rho
					rr = r[n:n+2]
					rho = density[n]
					## style
					style = dstyle(rho)
					## fill
					rvals=b.fill_between_r(rr, sty=style, npoints=1001, inf=50.)



def dstyle(rho, cm=plt.cm.Oranges, sty={}, ovr={}):
	"""
	Density rho in (0,1), output fill style.
	Since all Hayward have same l, all regions same scale.
	"""
	## style
	style = dict(fc='none', ec='none', lw=0, zorder=900)
	style.update(sty)
	## cnorm
	cmin, cmax = .1, .6
	cnorm = cmin + (cmax-cmin)*rho
	## color and alpha
	style.update(fc=cm(cnorm), alpha=.75)
	## override
	style.update(ovr)
	## return
	return style.copy()
		
			

def make_rlines(reglist, chainparams, l=.05, R=1., Tevap=None, sty={}):
	"""
	"""
	## params
	R = 1.*R
	l = 1.*l
	rcore = (R*l**2)**(1./3.)
	## general
	style = dict(ls='-', zorder=10, lw=.4, alpha=.2)
	style.update(sty)
	## l scale
	x = np.arange(0.,.5*R/l,.5)
	scale = l
	style.update(dict(c='c'))
	rlines(reglist, scale*x, sty=style, inf=10., npoints=2001.)
	## R scale
	x = np.arange(0.,20.01,.5)
	scale = R
	style.update(dict(c='m'))
	rlines(reglist, scale*x, sty=style, inf=25., npoints=1001.)
	## Tevap scale
	if Tevap is not None:
		x = np.arange(.8,25.01,.5)
		scale = Tevap
		style.update(dict(c='#0000f0'))
		rlines(reglist, scale*x, sty=style, inf=25., npoints=1001.)


def vticks(reglist, dv=1., inf1=100., inf2=50.+irr, sty={}):
	"""
	Remainder lets dv carry across regions.
	"""
	## init
	remainder = 0.
	## regions
	for reg in reglist:
		## blocks
		for b in reg.blocks:
			## last blocks only
			if b.j == len(reg.metfunc.rj)-2:
				## outer last blocks only
				if not np.isfinite(b.uvbounds['umin']):
					## get min and max
					vmin = np.max([b.uvbounds['vmin'], -inf2])
					vmax = np.min([b.uvbounds['vmax'],  inf2])
					## is it too small
					toosmall = False
					if vmax-vmin <= remainder:
						toosmall = True
					## if too small adjust remainder
					if toosmall==True:
						remainder = remainder - (vmax-vmin)
					## if big enough go
					if toosmall==False:
						## make array
						vv = remainder + np.arange(vmin, vmax, dv)
						## get new remainder
						remainder = dv - (vmax - vv[-1])
						## make and add curve
						crv = xh.curve()
						crv.uv = np.array([-inf1+0.*vv, 1.*vv])
						style = dict(markeredgecolor='k', alpha=0.3, ls='none', marker=(2,0,45), markersize=7, zorder=-100)
						style.update(sty)
						crv.sty.update(style)
						b.add_curves_uv([crv])




def uticks(reglist, du=1., inf1=100., inf2=50.+irr, sty={}):
	"""
	Remainder lets dv carry across regions.
	"""
	## init
	remainder = 0.
	## regions
	for reg in reglist:
		## blocks
		for b in reg.blocks:
			## last blocks only
			if b.j == len(reg.metfunc.rj)-2:
				## outer last blocks only
				if not np.isfinite(b.uvbounds['vmax']):
					## get min and max
					umin = np.max([b.uvbounds['umin'], -inf2])
					umax = np.min([b.uvbounds['umax'],  inf2])
					## is it too small
					toosmall = False
					if umax-umin <= remainder:
						toosmall = True
					## if too small adjust remainder
					if toosmall==True:
						remainder = remainder - (umax-umin)
					## if big enough go
					if toosmall==False:
						## make array
						uu = remainder + np.arange(umin, umax, du)
						## get new remainder
						remainder = du - (umax - uu[-1])
						## make and add curve
						crv = xh.curve()
						crv.uv = np.array([1.*uu, inf1+0.*uu])
						style = dict(markeredgecolor='k', alpha=.3, marker=(2,0,-45), ls='none', markersize=7, zorder=100)
						style.update(sty)
						crv.sty.update(style)
						b.add_curves_uv([crv])



def tticks(reglist, dt=1., inf1=1., inf2=50.+irr, sty={}):
	"""
	Remainder lets dt carry across regions.
	"""
	if reglist[0].metfunc.info['Type']=='Anti de Sitter':
		inf1 = 9.*reglist[0].metfunc.fparams['L']
		## init
		remainder = 0.
		## regions
		for reg in reglist:
			## blocks
			for b in reg.blocks:
				## last blocks only
				if b.j == len(reg.metfunc.rj)-2:
					## outer last blocks only
					if (not np.isfinite(b.uvbounds['vmax'])) or (not np.isfinite(b.uvbounds['umin'])):
						## get min and max
						umin = np.max([b.uvbounds['umin'], -inf2])
						umax = np.min([b.uvbounds['umax'],  inf2])
						vmin = np.max([b.uvbounds['vmin'], -inf2])
						vmax = np.min([b.uvbounds['vmax'],  inf2])
						tmin = 0.5*(vmin+umin)
						tmax = 0.5*(vmax+umax)
						## is it too small
						toosmall = False
						if tmax-tmin <= remainder:
							toosmall = True
						## if too small adjust remainder
						if toosmall==True:
							remainder = remainder - (tmax-tmin)
						## if big enough go
						if toosmall==False:
							print("HELLLOOOOOOOOO tticks")
							print(tmin, tmax, dt)
							## make array
							tt = remainder + np.arange(tmin, tmax, dt)
							## get new remainder
							remainder = dt - (tmax - tt[-1])
							## make and add curve
							crv = xh.curve()
							crv.tr = np.array([1.*tt,1.*inf1])
							style = dict(markeredgecolor='k', alpha=.3, marker=(2,0,90), ls='none', markersize=7, zorder=100)
							style.update(sty)
							crv.sty.update(style)
							b.add_curves_tr([crv])


## custom hatch
class CustomHatch(matplotlib.hatch.SmallFilledCircles):
    size = 0.1
    filled = True

    def __init__(self, hatch, density):
        density = 2
        self.num_rows = (hatch.count('c')) * density
        matplotlib.hatch.Circles.__init__(self, hatch, density)

## assign custom hatch
matplotlib.hatch._hatch_types.append(CustomHatch)



##############################

################# tools for drawing ########################################



def rlines(reglist, rvals, sty={}, npoints=2001, inf=25.):
	"""
	Add colorscaled lines of constant radius to region.
	Useful to check for matching.
	"""
	rvals = 1.*rvals
	for reg in reglist:
		for b in reg.blocks:
			for i in range(len(rvals)):
				if (b.rj[0]<rvals[i]) and (rvals[i]<b.rj[1]):
					style = dict(c='c', ls='-', lw=.5, zorder=1500)
					style.update(sty)
					rstarvals = 1.*b.master.metfunc.F(rvals[i:i+1])
					b.add_curves_uv(xh.cm.rstarlines_special_2(rstarvals, b.uvbounds, c=b.master.rparams['c'], sty=style, inf=1.*inf, npoints=1.*npoints))
	return reglist


def colorlines(reglist, rmin=1.1, rmax=25., dr=.5, sty={}, cm=plt.cm.hsv_r, npoints=2001, inf=25.):
	"""
	Add colorscaled lines of constant radius to region.
	Useful to check for matching.
	"""
	rvals = np.arange(rmin,rmax,dr)
	colvals = (rvals - np.min(rvals)) / (np.max(rvals) - np.min(rvals))
	cols = cm(colvals)
	for reg in reglist:
		for b in reg.blocks:
			for i in range(len(rvals)):
				if (b.rj[0]<rvals[i]) and (rvals[i]<b.rj[1]):
					style = dict(c=cols[i], ls='-', markersize=1, lw=.5, zorder=1500)
					style.update(sty)
					rstarvals = 1.*b.master.metfunc.F(rvals[i:i+1])
					b.add_curves_uv(xh.cm.rstarlines_special_2(rstarvals, b.uvbounds, c=b.master.rparams['c'], sty=style, inf=2.*inf, npoints=npoints))
	return reglist

def s0_lines(reglist, sty={}, npoints=1001, inf=50., eps=1e-24):
	"""
	"""
	for reg in reglist:
		for b in reg.blocks:
			x = np.array([(2.*reg.rparams['s0'] - eps)])
			style1 = dict(c='m', alpha=.5, lw=3, ls=':', zorder=5000)
			style2 = dict(c='c', alpha=.5, lw=3, ls=':', zorder=5000)
			style1.update(sty)
			style2.update(sty)
			b.add_curves_uv(xh.cm.uvlines( x, uv='uv', uvbounds=b.uvbounds, sty=style1, c=0., inf=inf, npoints=npoints))
			b.add_curves_uv(xh.cm.uvlines(-x, uv='uv', uvbounds=b.uvbounds, sty=style2, c=0., inf=inf, npoints=npoints))
	return reglist

##########################################################################################







"""
"""

import numpy as np

from xhorizon.diagram_tools.curve_class import curve
from xhorizon.diagram_tools.diagram_classes import region, block
from xhorizon.diagram_tools import curvemakers as cm






def EFreg(func, rparams={}, io=0, lr=0, basepoint=np.array([0.,0.]), boundary=True, uvgrid=False, rlines=True):
	"""
	Create an EF region with given metfunc.
		basepoint_uvdl gives coordinates of the outermost horizon vertex.
		io makes the region ingoing (0) or outgoing (1)
		lr makes the region left-directed or right-directed
	"""
	## get func params
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## get region params
	rparam = dict(c=0.,s0=10.)
	rparam.update(rparams)
	## create region
	reg = region(func, rparams=rparam)
	## get basepoint
	x0, y0 = basepoint
	udl0, vdl0 = 0.5*(y0-x0), 0.5*(y0+x0)
	## add blocks
	## ingoing io=0 lr=0
	if io==0 and lr==0:
		for j in range(N+1):
			cdlu = udl0 - 0.5 + (N-j)
			cdlv = vdl0 + 0.5
			reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=func.sgnf(j), epsv=1.))
	## ingoing io=0 lr=1
	if io==0 and lr==1:
		for j in range(N+1):
			cdlu = udl0 - 0.5
			cdlv = vdl0 - 1.5 + (j)
			reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=1., epsv=func.sgnf(j) ))
	## outgoing io=1 lr=0
	if io==1 and lr==0:
		for j in range(N+1):
			cdlu = udl0 - 1.5 + (j)
			cdlv = vdl0 - 0.5
			reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-func.sgnf(j), epsv=-1.))
	## outgoing io=1 lr=1
	if io==1 and lr==1:
		for j in range(N+1):
			cdlu = udl0 + 0.5
			cdlv = vdl0 - 0.5 + (N-j)
			reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-1., epsv=-func.sgnf(j) ))
	## add boundary
	if boundary==True:
		for b in reg.blocks:
			b.add_curves_uv(cm.block_boundary(b))
	## add grid
	if uvgrid==True:
		for b in reg.blocks:
			uvvals = np.arange(-30.,30., 0.5)
			b.add_curves_uv(cm.uvlines(uvvals, uv='uv', sty=dict(c='0.5')))
	## add default rlines
	if rlines==True:
		for b in reg.blocks:
			b.add_curves_tr(cm.default_rlines(reg.metfunc))
	## return
	return reg


def MAXreg(func, rparams={}, boundary=True, uvgrid=False, rlines=True):
	"""
	Create a maximally extended region with given metfunc.
	"""
	## get N and eps0
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## send to relevant subroutine
	if N==0:
		reg = MAXreg0(func, rparams=rparams, boundary=boundary, uvgrid=uvgrid, rlines=rlines)
	elif N==1 and eps0==-1.:
		reg = MAXreg1a(func, rparams=rparams, boundary=boundary, uvgrid=uvgrid, rlines=rlines)
	elif N==1 and eps0== 1.:
		reg = MAXreg1b(func, rparams=rparams, boundary=boundary, uvgrid=uvgrid, rlines=rlines)
	elif N == 2 and eps0== 1.:
		reg = MAXreg2a(func, rparams=rparams, boundary=boundary, uvgrid=uvgrid, rlines=rlines)
	elif N == 2 and eps0==-1.:
		reg = MAXreg2b(func, rparams=rparams, boundary=boundary, uvgrid=uvgrid, rlines=rlines)
	else:
		reg = None
		print("\nNo existing subroutine for MAXreg at this value of N and eps0.\n")
	## format
	## add boundary
	if boundary==True:
		for b in reg.blocks:
			b.add_curves_uv(cm.block_boundary(b))
	## add grid
	if uvgrid==True:
		for b in reg.blocks:
			uvvals = np.arange(-30.,30., 0.5)
			b.add_curves_uv(cm.uvlines(uvvals, uv='uv', sty=dict(c='0.5')))
	## add default rlines
	if rlines==True:
		for b in reg.blocks:
			b.add_curves_tr(cm.default_rlines(reg.metfunc))
	## return
	return reg


def MAXreg2a(func, rparams={}, boundary=True, uvgrid=False, rlines=True):
	"""
	Create a maximally extended region with given metfunc with two horizons.
	"""
	## get func params
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## make sure there are exactly zero horizons
	if N != 2:
		return None
	## get region params
	rparam = dict(c=0.,s0=10.)
	rparam.update(rparams)
	## create region
	reg = region(func, rparams=rparam)
	## add blocks
	## set basepoint
	udl0, vdl0 = 0., 0.
	## ingoing io=0 lr=0
	for j in [0,1,2]:
		cdlu = udl0 - 0.5 + (N-j)
		cdlv = vdl0 + 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=func.sgnf(j), epsv=1.))
	## ingoing io=0 lr=1
	for j in [0]:
		cdlu = udl0 - 0.5
		cdlv = vdl0 - 1.5 + (j)
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=1., epsv=func.sgnf(j) ))
	## outgoing io=1 lr=0
	for j in [0,1,2]:
		cdlu = udl0 - 1.5 + (j)
		cdlv = vdl0 - 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-func.sgnf(j), epsv=-1.))
	## outgoing io=1 lr=1
	for j in [0]:
		cdlu = udl0 + 0.5
		cdlv = vdl0 - 0.5 + (N-j)
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-1., epsv=-func.sgnf(j) ))
	## return
	return reg

def MAXreg2b(func, rparams={}, boundary=True, uvgrid=False, rlines=True):
	"""
	Create a maximally extended region with given metfunc with two horizons.
	"""
	## get func params
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## make sure there are exactly zero horizons
	if N != 2:
		return None
	## get region params
	rparam = dict(c=0.,s0=10.)
	rparam.update(rparams)
	## create region
	reg = region(func, rparams=rparam)
	## add blocks
	## set basepoint
	udl0, vdl0 = 0., 0.
	## ingoing io=0 lr=0
	for j in [0,1,2]:
		cdlu = udl0 - 0.5 + (N-j)
		cdlv = vdl0 - 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=func.sgnf(j), epsv=1.))
	## ingoing io=0 lr=1
	for j in [0]:
		cdlu = udl0 + 0.5
		cdlv = vdl0 - 1.5 + (j)
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=1., epsv=func.sgnf(j) ))
	## outgoing io=1 lr=0
	for j in [0,1,2]:
		cdlu = udl0 - 1.5 + (j)
		cdlv = vdl0 + 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-func.sgnf(j), epsv=-1.))
	## outgoing io=1 lr=1
	for j in [0]:
		cdlu = udl0 - 0.5
		cdlv = vdl0 - 0.5 + (N-j)
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-1., epsv=-func.sgnf(j) ))
	## return
	return reg

def MAXreg1a(func, rparams={}, boundary=True, uvgrid=False, rlines=True):
	"""
	Create a maximally extended region with given metfunc with two horizons.
	"""
	## get func params
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## make sure there are exactly zero horizons
	if N != 1:
		return None
	## get region params
	rparam = dict(c=0.,s0=10.)
	rparam.update(rparams)
	## create region
	reg = region(func, rparams=rparam)
	## add blocks
	## set basepoint
	udl0, vdl0 = 0., 0.
	## ingoing io=0 lr=0
	for j in [0,1]:
		cdlu = udl0 - 0.5 + (N-j)
		cdlv = vdl0 + 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=func.sgnf(j), epsv=1.))
	## outgoing io=1 lr=0
	for j in [0,1]:
		cdlu = udl0 - 0.5 + (j)
		cdlv = vdl0 - 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-func.sgnf(j), epsv=-1.))
	## return
	return reg




def MAXreg1b(func, rparams={}, boundary=True, uvgrid=False, rlines=True):
	"""
	Create a maximally extended region with given metfunc with two horizons.
	"""
	## get func params
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## make sure there are exactly zero horizons
	if N != 1:
		return None
	## get region params
	rparam = dict(c=0.,s0=10.)
	rparam.update(rparams)
	## create region
	reg = region(func, rparams=rparam)
	## add blocks
	## set basepoint
	udl0, vdl0 = 0., 0.
	## ingoing io=0 lr=0
	for j in [0,1]:
		cdlu = udl0 - 0.5 + (N-j)
		cdlv = vdl0 - 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=func.sgnf(j), epsv=1.))
	## outgoing io=1 lr=0
	for j in [0,1]:
		cdlu = udl0 - 0.5 + (j)
		cdlv = vdl0 + 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=-func.sgnf(j), epsv=-1.))
	## return
	return reg





def MAXreg0(func, rparams={}, boundary=True, uvgrid=False, rlines=True):
	"""
	Create a maximally extended region with given metfunc with two horizons.
	"""
	## get func params
	N = len(func.rj)-2
	eps0 = func.sgnf(0)
	## make sure there are exactly zero horizons
	if N != 0:
		return None
	## get region params
	rparam = dict(c=0.,s0=10.)
	rparam.update(rparams)
	## create region
	reg = region(func, rparams=rparam)
	## add blocks
	## set basepoint
	udl0, vdl0 = 0., 0.
	## ingoing io=0 lr=0
	for j in [0]:
		cdlu = udl0 - 0.5 + (N-j)
		cdlv = vdl0 + 0.5
		reg.add_block(j, bparams=dict(cdlu=cdlu, cdlv=cdlv, epsu=func.sgnf(j), epsv=1.))
	## return
	return reg





from xhorizon.tortoise import math_util
from xhorizon.tortoise import metfunc_class
from xhorizon.tortoise import metfunc_factory
from xhorizon.tortoise import metfunc_tests
from xhorizon.tortoise import tortoise




"""
This module defines a class used for evaluating coordinate transformations at null shell junctions.
"""


import numpy as np
import xhorizon.shell_junction.interpolators as interp
from xhorizon.shell_junction.helpers import *



class passive_slice:

	"""
	Similar to active slice, except instead of setting new U(udl) and V(vdl) function, this one merely reads the
	existing ones. This is used to obtain functions to act as input for active slice U0,V0 params.
	"""

	def __init__(self, reg, ublocks=[], vblocks=[], r0=np.nan, u0=np.nan, v0=np.nan, mu=0., r_refs=[]):

		## process and store input values
		self.reg = reg
		self.ublocks, self.vblocks = bcon(reg,ublocks), bcon(reg,vblocks)
		self.r0, self.u0, self.v0 = set_ruv0(reg, r0=r0, u0=u0, v0=v0)
		self.mu = 1.*float(mu)

		## get r, uvdl ref arrays
		self.r = get_r_ref(self.reg, r_refs, self.r0)
		self.r, self.uvdl_u0, self.uvdl_v0 = uvdl_of_r_at_uv0(self.r, self.reg, ublocks=self.ublocks, vblocks=self.vblocks, u0=self.u0, v0=self.v0)
		
		## get UV ref arrays
		self.UV_u0 = reg.UV_of_uvdl(self.uvdl_u0)
		self.UV_v0 = reg.UV_of_uvdl(self.uvdl_v0)

	"""
	Coordinate transformation methods.
	"""

	def U_of_udl_at_v0(self, udl):
		"""
		Evaluate the function U(udl,vdl0) by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		"""
		return interp.interp_with_smooth_extrap(udl, self.uvdl_v0[0], self.UV_v0[0], mu=self.mu)

	def V_of_vdl_at_u0(self, vdl):
		"""
		Evaluate the function V(udl0,vdl) by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		"""
		return interp.interp_with_smooth_extrap(vdl, self.uvdl_u0[1], self.UV_u0[1], mu=self.mu)


	def U_of_r_at_v0(self, r):
		"""
		Evaluate the function U(r) at v0 by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		Setting mu to nan because this should not be extrapolated.
		"""
		return interp.interp_with_smooth_extrap(r, self.r, self.UV_v0[0], mu=np.nan)

	def V_of_r_at_u0(self, r):
		"""
		Evaluate the function V(r) at u0 by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		Setting mu to nan because this should not be extrapolated.
		"""
		return interp.interp_with_smooth_extrap(r, self.r, self.UV_u0[1], mu=np.nan)


	def udl_of_U_at_v0(self, U):
		"""
		Untested.
		Inverse of U(udl).
		"""
		return interp.interp_with_smooth_extrap(U, self.UV_v0[0], self.uvdl_v0[0], mu=self.mu)

	def vdl_of_V_at_u0(self, V):
		"""
		Untested.
		Inverse of V(vdl).
		"""
		return interp.interp_with_smooth_extrap(V, self.UV_u0[1], self.uvdl_u0[1], mu=self.mu)





"""
This module contains helpers to help with evap code.
"""

import numpy as np
import matplotlib.pyplot as plt
import copy

import xhorizon as xh
from xhorizon.shell_junction import interpolators as interp

### constants ###
irr = .5 * ((np.pi+np.e)%1.)

################ tools for choosing junction corner radius ###############



def get_rinf_uv0(reglist, v0=[]):
	"""
	Given a list of regions, assume that `outermost' region has index [-1].
	Assume r->inf as u->-inf.
	Find the r0 value corresponding to (v=v0 and u=2.*s0) in each region.
	Take minimum r0 over reglist.
	Return a value slightly less than r0.
	"""
	r0 = np.nan * np.ones(len(reglist))
	for i in range(len(reglist)):
		reg = reglist[i]
		for b in [reg.blocks[-1]]:
			u = np.array([-2.*reg.rparams['s0']])
			v = np.array([v0[i]])
			uv = np.array([u,v])
			r0[i] = b.tr_of_uv(uv)[1]
	r0 = .85*np.min(r0)
	## bug fix for AdS
	if not np.isfinite(r0):
		if 'L' in reglist[0].metfunc.fparams.keys():
			r0=1.9*reglist[0].metfunc.fparams['L']
	## return
	return r0


###############################################################################
				

############# tools to rescale all coordinates at the end ###################

def UVcompose1(reg, fU=None, fV=None):
	"""
	Take the final UV coords for region and compose with a function in each null direction.
	"""
	reg2 = copy.deepcopy(reg)
	if not fU==None:
		reg.U_of_udl = lambda x: fU(reg2.U_of_udl(x))
	if not fV==None:
		reg.V_of_vdl = lambda x: fU(reg2.V_of_vdl(x))
	return reg

def UVcompose(reglist, fU=None, fV=None):
	"""
	Take the final UV coords for region and compose with a function in each null direction.
	Takes list input.
	"""
	for reg in reglist:
		reg = UVcompose1(reg, fU=fU, fV=fV)
	return reglist

def get_uvdl_of_UV(reg):
	"""
	Get current udl_of_U and vdl_of_V for a region by evaluation and interpolation.
	"""
	## copy region
	reg2 = copy.deepcopy(reg)
	## get x_ref arrays
	ss = np.linspace(-4,4,50001)
	uudl = 1.*ss
	vvdl = 1.*ss
	## get y_ref arrays
	UU = reg.U_of_udl(uudl)
	VV = reg.V_of_vdl(vvdl)
	## make functions
	udl_of_U = lambda U: interp.interp_with_smooth_extrap(U, UU, uudl, mu=0.)
	vdl_of_V = lambda V: interp.interp_with_smooth_extrap(V, VV, vvdl, mu=0.)
	## return
	return udl_of_U, vdl_of_V











############################################################################


############## plot reglist ####################

def rgp(reglist):
	"""
	Plot all regions in reglist.
	"""
	## lines
	xh.evap.colorlines(reglist, sty=dict(alpha=0.5, c='k'))
	#xh.evap.uv_lines(reglist, uv='u', sty=dict())
	xh.evap.boundarylines(reglist, sty=dict(), npoints=5001)
	xh.evap.s0_lines(reglist, sty=dict(lw=3))
	## figure
	plt.figure(figsize=(6,6))
	plt.axes([.1, .1, .8, .8])
	plt.xlim(-3,3)
	plt.ylim(-3,3)
	plt.gca().set_aspect('equal')
	## plot
	for reg in reglist:
		reg.rplot()
	## fill
	xh.evap.fill_by_R_2(reglist, col='r', amax=.2)

#############################################################


################# tools for drawing ########################################

def boundarylines(reglist, npoints=5001, sty={}):
	"""
	Add boundary lines to regions in reglist.
	"""
	for reg in reglist:
		for b in reg.blocks:
			style = dict(lw=0.9, c='0.5', zorder=2000)
			style.update(sty)
			b.add_curves_uv(xh.cm.block_boundary_2(b, sty=style, npoints=npoints))


def colorlines(reglist, rmin=0.05, rmax=5., dr=.2, sty={}, cm=plt.cm.hsv_r, npoints=2001, inf=25.):
	"""
	Add colorscaled lines of constant radius to region.
	Useful to check for matching.
	"""
	rvals = np.arange(rmin,rmax,dr)
	colvals = (rvals - np.min(rvals)) / (np.max(rvals) - np.min(rvals))
	cols = cm(colvals)
	for reg in reglist:
		for b in reg.blocks:
			for i in range(len(rvals)):
				if (b.rj[0]<rvals[i]) and (rvals[i]<b.rj[1]):
					style = dict(c=cols[i], ls='-', markersize=1, lw=.5, zorder=1500)
					style.update(sty)
					rstarvals = 1.*b.master.metfunc.F(rvals[i:i+1])
					b.add_curves_uv(xh.cm.rstarlines_special_2(rstarvals, b.uvbounds, c=b.master.rparams['c'], sty=style, inf=2.*inf, npoints=npoints))
	return reglist

def s0_lines(reglist, sty={}, npoints=1001, inf=50., eps=1e-24):
	"""
	"""
	for reg in reglist:
		for b in reg.blocks:
			x = np.array([(2.*reg.rparams['s0'] - eps)])
			style1 = dict(c='m', alpha=.5, lw=3, ls=':', zorder=5000)
			style2 = dict(c='c', alpha=.5, lw=3, ls=':', zorder=5000)
			style1.update(sty)
			style2.update(sty)
			b.add_curves_uv(xh.cm.uvlines( x, uv='uv', uvbounds=b.uvbounds, sty=style1, c=0., inf=inf, npoints=npoints))
			b.add_curves_uv(xh.cm.uvlines(-x, uv='uv', uvbounds=b.uvbounds, sty=style2, c=0., inf=inf, npoints=npoints))
	return reglist


def uv_lines(reglist, uv='uv', sty={}, npoints=1001, inf=50., eps=1e-24):
	"""
	"""
	for reg in reglist:
		for b in reg.blocks:
			smin, smax, ds = -5., 25., 1.
			vals = np.arange(smin, smax, ds)
			cm = plt.cm.gist_rainbow
			cv = np.linspace(0,1,len(vals))
			for i in range(len(vals)):
				style1 = dict(c=cm(cv[i]), lw=.8, ls='-', zorder=6000)
				style1.update(sty)
				b.add_curves_uv(xh.cm.uvlines([vals[i]], uv=uv, uvbounds=b.uvbounds, sty=style1, c=0., inf=inf, npoints=npoints))
	return reglist

##########################################################################################




######################### tools for filling #############################################

def fillcols_by_R(reglist):
	"""
	Get fill color values based on radius.
	"""
	## get rvals
	Rvals = np.zeros(len(reglist))
	for i, reg in enumerate(reglist):
		if 'R' in reg.metfunc.fparams.keys():
			x = reg.metfunc.fparams['R']
			Rvals[i] = x
	## get colvals
	norm = np.max(Rvals)
	colvals = 0.2 + 0.7 * Rvals / norm
	#print("Rvals = %s"%(Rvals))
	#print("colvals = %s"%(colvals))
	## return
	return colvals

def alphas_by_R_2(reglist, amax=.9):
	"""
	Get fill color values based on radius.
	"""
	## rvals
	Rvals = np.zeros(len(reglist))
	for i, reg in enumerate(reglist):
		if 'R' in reg.metfunc.fparams.keys():
			x = reg.metfunc.fparams['R']
			Rvals[i] = 1.*x
	## log
	Rvals = np.log(1.+Rvals)
	## alphas
	amin, amax = 0., amax
	avals = amin + (amax-amin)*(Rvals/np.max(Rvals))
	## return
	return 1.*avals


def fill_by_R(reglist, cm=plt.cm.prism):
	colvals = fillcols_by_R(reglist)
	for i in range(len(reglist)):
		reg = reglist[i]
		col = cm(colvals[i])
		for b in reg.blocks:
			sty = dict(fc=col, alpha=.2)
			b.fill(sty=sty)

def fill_by_R_2(reglist, col='r', amax=.9):
	##
	alphas = alphas_by_R_2(reglist, amax=amax)
	for i in range(len(reglist)):
		reg = reglist[i]
		for b in reg.blocks:
			sty = dict(fc=col, alpha=alphas[i])
			b.fill(sty=sty)

############################################################################################


################################ checkers ################################

def evaporation_input_good(R, du, dv):
	"""
	"""
	## initialize
	good = True
	## check
	if not (len(R)==len(du) and len(R)==len(dv)):
		print("MISMATCHED INPUT ARRAY LENGTHS")
		good = False
	if not np.all(R>0.):
		print("RADIUS MUST BE POSITIVE")
		good = False
	## return
	return good

def check_uvr(reglist):
	"""
	"""
	## first values
	iu = [0,0]
	iv = [1,1]
	uu = [reglist[0].blocks[-1].uvbounds['umin'], reglist[0].blocks[-1].uvbounds['umax']]
	vv = [reglist[1].blocks[-1].uvbounds['vmin'], reglist[1].blocks[-1].uvbounds['vmax']]
	## middle values
	for i in range(len(reglist)):
		reg = reglist[i]
		## uu
		a, b = reg.blocks[-1].uvbounds['umin'], reg.blocks[-1].uvbounds['umax']
		if np.isfinite(a) and np.isfinite(b):
			uu += [a,b]
			iu += [i,i]
		## vv
		a, b = reg.blocks[-1].uvbounds['vmin'], reg.blocks[-1].uvbounds['vmax']
		if np.isfinite(a) and np.isfinite(b):
			vv += [a,b]
			iv += [i,i]
	## last values
	iu += [-1,-1]
	iv += [-1,-1]
	uu += [reglist[-1].blocks[-1].uvbounds['umin'], reglist[-1].blocks[-1].uvbounds['umax']]
	vv += [reglist[-1].blocks[-1].uvbounds['vmin'], reglist[-1].blocks[-1].uvbounds['vmax']]
	## radius
	RRu = []
	for i in iu:
		if 'R' in reglist[i].metfunc.fparams.keys():
			RRu += [reglist[i].metfunc.fparams['R']]
		else:
			RRu += [0.]
	## radius
	RRv = []
	for i in iv:
		if 'R' in reglist[i].metfunc.fparams.keys():
			RRv += [reglist[i].metfunc.fparams['R']]
		else:
			RRv += [0.]
	## radius
	rru = []
	for k in range(len(iu)):
		rru += [reglist[iu[k]].blocks[-1].r_of_uv(np.array([[uu[k]],[vv[k]]]))[0]]
	## radius
	rrv = []
	for k in range(len(iv)):
		rrv += [reglist[iv[k]].blocks[-1].r_of_uv(np.array([[uu[k]],[vv[k]]]))[0]]
	## array
	iu,iv,uu,vv,rru,rrv,RRu,RRv = np.array(iu),np.array(iv),np.array(uu),np.array(vv),np.array(rru),np.array(rrv),np.array(RRu),np.array(RRv)
	## print
	print("\n")
	print("CHECK_UVR:")
	print("iu  = %r"%(iu))
	print("iv  = %r"%(iv))
	print("uu  = %r"%(uu))
	print("vv  = %r"%(vv))
	print("rru = %r"%(rru))
	print("rrv = %r"%(rrv))
	print("RRu = %r"%(RRu))
	print("RRv = %r"%(RRv))
	print("rru/RRu = " + repr(rru/RRu))
	print("\n")


############################################################################


######### get mass parameters #######
def getmm(reglist):
	## initialize masses
	m = np.zeros(len(reglist))
	## loop
	for i in range(len(reglist)):
		fp = reglist[i].metfunc.fparams
		if 'R' in fp.keys():
			m[i] = 0.5*fp['R']
	## return
	return 1.*m
#############################


############## split region into abcd ####################

def split_reg_abcd(reg, abcd='abcd', u0=0., v0=0.):
	"""
	"""
	##
	reglist = []
	##
	for x in abcd:
		reglist += [xh.cornermask.EFreg(copy.deepcopy(reg), abcd=x, u0=u0, v0=v0)]
	## return
	return reglist

#############################################################


#################### other #########################################


def sliceprint(pslice, aslice, Rh1, Rh2):
	print("\n")
	print("                          %22r, %22r, %22r, %22r, %22r"%('Rh', 'r', 'r-Rh', 'u', 'v'))
	print("PSlice:")
	print("Rh1, r1, r1/Rh1, u1, v1 = %22r, %22r, %22r, %22r, %22r"%(Rh1, pslice.r0, pslice.r0-Rh1, pslice.u0, pslice.v0))
	print("ASlice:")
	print("Rh2, r2, r2/Rh2, u2, v2 = %22r, %22r, %22r, %22r, %22r"%(Rh2, aslice.r0, aslice.r0-Rh2, aslice.u0, aslice.v0))
	print("\n")


def mp(matchmode):
	matchpop = 'uvr0'
	for s in matchmode:
		matchpop = matchpop.replace(s,'')
	return matchpop

#############################################################





"""
This module defines the class
	metfunc
which is used to encode information about the SSS metric function f(r).
"""

import numpy as np



class metfunc:

	"""
	The metfunc class contains information derived from the SSS metric function f(r).
	The default values correspond to a Minkowski spacetime with f(r)=1.
	"""

	def __init__(self):
		## metric function parameters
		self.fparams = dict()
		## tortoise function parameters
		self.Fparams = dict(eps=1e-9, inf=25.)
		## radii r_j
		self.rj = np.array([0., self.Fparams['inf']])
		## slopes k_j
		self.kj = np.array([0., 0.])
		## printable info dict
		self.info = {
		'Type' : "Minkowski",
		'Metric Function' : r"$f(r) = 1$",
		'Parameters' : "",
		}
		## interpolation reference values for F(r)
		self.r_ref = np.array([])
		self.rstar_ref = np.array([])
		## first trapped interval (0 or 1)
		self.first_trapped_j = 1

	## metric function f(r)
	def f(self, r):
		return 1. + 0. * r

	## tortoise function F(r)
	def F(self, r):
		rstar = 1. * r
		return rstar

	## inverse tortoise function Finv_j(rstar)
	def Finv(self, j, rstar):
		## j and rstar are equal length arrays
		## if interval number is invalid return nan
		r = np.nan * rstar
		## interval I_0
		mask = (j==0)
		r[mask] = rstar[mask]
		## return
		return r

	## which interval based on rj
	def winterval(self, r):
		j = 0. * r
		return j

	## sign of f in interval I_j
	def sgnf(self, j):
		sgn = np.nan
		N = len(self.rj) - 2
		## valid index?
		if j in range(N+1):
			## is sgnf positive?
			plus = ( int(j) + int(self.first_trapped_j) ) % 2
			## fill value
			if plus==0:
				sgn = -1.
			if plus==1:
				sgn = 1.
		## return
		return sgn












"""
This module provides routines for setting the matplotlib style and creating new figures in the proper style.
"""

import matplotlib as mpl
import matplotlib.pyplot as plt



def pubrc(tex=False):
	## if matplotlib version 2, use classic style
	if int(mpl.__version__.split('.')[0]) > 1:
		plt.style.use('classic')
	## set default rc params for diagrams
	plt.rcParams['font.family'] = 'serif'
	## latex settings
	if tex==True:
		plt.rcParams['text.usetex'] = True
		plt.rcParams['font.family'] = 'serif'
		plt.rcParams['font.serif'] = []
		plt.rcParams['text.latex.preamble'] = [r'\renewcommand{\seriesdefault}{\bfdefault}']
	## line params
	plt.rcParams['lines.linewidth'] = 0.5
	## font params
	plt.rcParams['font.size'] = 10
	## legend params
	plt.rcParams['legend.fontsize'] = 'medium'



def newfig(sqfig=4, sqaxis=4, tilde=False, tex=False):
	## set rc params
	pubrc(tex=tex)
	## new figure
	plt.figure(figsize=(sqfig,sqfig))
	## new axes
	left, bottom, right, top = .105, .105, .97, .97
	width, height = right-left, top-bottom
	plt.axes([left, bottom, width, height])
	## format axes
	xpad = 4
	ypad = 0
	plt.xlabel(r'$V-U$', labelpad=xpad)
	plt.ylabel(r'$V+U$', labelpad=ypad)
	## tilde instead of cap?
	if tilde==True:
		plt.xlabel(r'$\tilde{v}-\tilde{u}$', labelpad=xpad)
		plt.ylabel(r'$\tilde{v}+\tilde{u}$', labelpad=ypad)
	## range
	sqaxis = sqaxis * 1.1
	plt.xlim(-sqaxis,sqaxis)
	plt.ylim(-sqaxis,sqaxis)











"""
Helpers for evaluating coordinate transformations at null shell junctions.
"""


import numpy as np




"""
Helpers.
"""

def set_ruv0(reg, r0=np.nan, u0=np.nan, v0=np.nan):
	"""
	Self-consistently fill all three values based on input.
	For a shell junction, r0 should be nan and (u0 xor v0) should be finite.
	For a corner junction, r0 should be finite and (u0 xor v0) should be finite.
	Region arg is needed to provide c and F(r).
	If invalid, warn and return all nan values.
	"""
	## initialize
	ruv0 = [np.nan, np.nan, np.nan]
	F = reg.metfunc.F
	c = float(reg.rparams['c'])
	r0, u0, v0 = 1.*float(r0), 1.*float(u0), 1.*float(v0)
	## only go if (u0 xor v0) is finite
	if np.isfinite(u0)^np.isfinite(v0):
		## shell
		if not np.isfinite(r0):
			ruv0 = [np.nan, u0, v0]
		## corner
		if np.isfinite(r0):
			## u0 given
			if np.isfinite(u0):
				v0 = u0 + 2.*(F(r0)-c)
			## v0 given
			if np.isfinite(v0):
				u0 = v0 - 2.*(F(r0)-c)	
			## set values
			ruv0 = [r0, u0, v0]
	## return
	return ruv0




def uvdl_of_r_at_uv0(r, reg, ublocks=[], vblocks=[], u0=np.nan, v0=np.nan):
	"""
	Calculate uvdl of r at fixed u0 and v0 at block points. Slices in both directions are calculated.
		r = An array of r values at which to evaluate. Non block points return nan.
		reg = A region in which to evaluate.
		ublocks =  A list of integer indices for blocks in reg.blocks. The u0 slice is evaluated at block points in these blocks only.
		vblocks =  A list of integer indices for blocks in reg.blocks. The v0 slice is evaluated at block points in these blocks only.
			Alternately, block args can be provided directly as a list of blocks (bcon converter handles this).
			Block args necessary bc radius alone can be ambiguous in extended regions. 
		u0 = Fixed value of u0 at which the slice is evaluated.
		v0 = Fixed value of v0 at which the slice is evaluated.
	Returns (r, uvdl_u0, uvdl_v0).
	"""
	## initialize
	u0, v0 = 1.*float(u0), 1.*float(v0)
	uvdl_u0 = np.nan*np.ones((2,len(r)))
	uvdl_v0 = np.nan*np.ones((2,len(r)))
	u,v = np.nan*r, np.nan*r
	rstar = reg.metfunc.F(1.*r)
	c = 1.*float(reg.rparams['c'])
	## convert block args
	ublocks = bcon(reg, ublocks)
	vblocks = bcon(reg, vblocks)
	## u0 slice
	for b in ublocks:
		## mask points outside block and calculate
		mask = np.logical_and.reduce([b.rj[0]<r, r<b.rj[1]])
		u[mask] = u0 + 0.*rstar[mask]
		v[mask] = u0 + 2.*(rstar[mask]-c)
		uvdl_u0[:,mask] = b.uvdl_of_uv(np.array([u[mask],v[mask]]))
	## v0 slice
	for b in vblocks:
		## mask points outside block and calculate
		mask = np.logical_and.reduce([b.rj[0]<r, r<b.rj[1]])
		v[mask] = v0 + 0.*rstar[mask]
		u[mask] = v0 - 2.*(rstar[mask]-c)
		uvdl_v0[:,mask] = b.uvdl_of_uv(np.array([u[mask],v[mask]]))
	## return
	return r, uvdl_u0, uvdl_v0



def U_of_udl_at_v0(r, udl, U0):
	"""
	Implements the function U(udl) = U0( r(udl,vdl0) ) at v=v0, where U0(r) is a function of radius.
	If U0==None, shell mode is activated, and the function reduces to U(udl) = udl.
	Inputs:
		r = An array of radius values at points along v=v0.
		udl = An array of udl values corresponding to the points in r.
		U0 = For corner mode, a monotonic function U0(r). If U0==None, shell mode activates.
	Returns:
		U_v0 = An array of U values corresponding the points in r at v=v0.
	"""
	## initialize
	U_v0 = np.nan*r
	## corner mode
	if not U0==None:
		U_v0 = U0(1.*r)
	## shell mode
	if U0==None:
		U_v0 = 1.*udl
	## return
	return 1.*U_v0


def V_of_vdl_at_u0(r, vdl, V0):
	"""
	Implements the function V(vdl) = V0( r(udl0,vdl) ) at u=u0, where V0(r) is a function of radius.
	If V0==None, shell mode is activated, and the function reduces to V(vdl) = vdl.
	Inputs:
		r = An array of radius values at points along u=u0.
		vdl = An array of vdl values corresponding to the points in r.
		V0 = For corner mode, a monotonic function V0(r). If V0==None, shell mode activates.
	Returns:
		V_u0 = An array of V values corresponding the points in r at u=u0.
	
	This function is identical to U_of_udl_at_v0. To avoid redundancy, this one just calls the other.
	They are kept separate only to make the code where these are called more readable.
	"""
	## return
	return U_of_udl_at_v0(r, vdl, V0)



def bcon(reg, blocks):
	"""
	Block list converter.
	If blocks is a list of integers, convert to a list of blocks using reg.blocks.
	Else, do nothing.
	Return list of blocks.
	"""
	## go
	if len(blocks)>0 and type(blocks[0])==int:
		blocks = [ reg.blocks[blocks[i]] for i in range(len(blocks)) ]
	## return
	return blocks



def dtr_calc(f0):
	"""
	Given the values f(r0) in three of the regions, return the last one.
	The one to be calculated is specified by nan. 
	Order is [A,B,C,D].
	"""
	f0 = 1.*f0
	if not np.isfinite(f0[0]):
		f0[0] = f0[2]*f0[3] / f0[1]
	if not np.isfinite(f0[1]):
		f0[1] = f0[2]*f0[3] / f0[0]
	if not np.isfinite(f0[2]):
		f0[2] = f0[0]*f0[1] / f0[3]
	if not np.isfinite(f0[3]):
		f0[3] = f0[0]*f0[1] / f0[2]
	return f0


def schwarz_dtr_masses(m, r0=1.):
	"""
	Given three of four masses, return dtr compatible remaining mass.
	"""
	f0 = 1. - 2.*m/r0
	f0 = dtr_calc(f0)
	m = r0*(1.-f0)/2.
	return m



def slicecheck(sl,reg):
	if (np.abs(sl.u0)>=2.*reg.rparams['s0']) or (np.abs(sl.v0)>=2.*reg.rparams['s0']):
		print("Error: u0 or v0 out of range.")
		print(sl.u0, sl.v0)
		sl, reg = None, None
	return sl, reg


def get_r_ref(reg, r_refs, r0):
	"""
	"""
	## init
	rrlist = []
	## add reg metfunc r_ref
	rrlist += [1.*reg.metfunc.r_ref]
	## add r_refs values
	rrlist += [1.*rrr for rrr in r_refs]
	## add values near r0
	a, npoints = 1e-21, 500
	ss = np.linspace(0., np.log(a), npoints)
	#rrlist += [r0 + np.exp(1.*ss), r0 - np.exp(1.*ss)]
	## concatenate
	rr = np.concatenate(rrlist)
	## return
	return 1.*rr













def demo():
	import numpy as np
	import matplotlib.pyplot as plt
	import xhorizon as xh
	from xhorizon.evap.massplot import massplot
	from xhorizon.evap import helpers

	"""
	"""

	## setup figure #################
	## rc
	plt.rcParams['font.size'] = 7
	plt.rcParams['axes.linewidth'] = .4
	plt.rcParams['xtick.major.size'] = 2
	plt.rcParams['ytick.major.size'] = 2
	plt.rcParams['lines.solid_capstyle'] = 'butt'
	plt.rcParams['lines.dash_capstyle'] = 'butt'
	## figure size
	fig_width  = 2.5  ## inches
	fig_aspect = 1.68   ## height/width
	fig_height = fig_width * fig_aspect
	## axis size
	width  = 0.86   ## x fraction
	aspect = 1.84   ## absolute height/width
	height = aspect * width / fig_aspect    ## y fraction
	## axis loc
	left, bottom = .12, .05     ## x fraction, y fraction
	## define figure and axes
	plt.figure(1,figsize=(fig_width,fig_height))
	ax1 = plt.axes([left,bottom,width,height], aspect=1.)
	ax = [ax1]
	## format axes
	for axx in ax:
		## set axis
		plt.sca(axx)
		## labels
		plt.xlabel('$V-U$', labelpad=-7)
		plt.ylabel('$V+U$', labelpad=-4)
		## ticks
		plt.xticks([-1,1])
		plt.yticks([-1,1])
		## lims
		sq = 2.4
		x0 = -1.2
		y0 = 0.
		plt.xlim(x0,x0+sq)
		plt.ylim(y0-0.5*aspect*sq,y0+0.5*aspect*sq)
	################################


	## save
	path = "temp-figs/demo"
	sfp = dict(dpi=800)
	temp_only = False

	## draw
	draw = True

	## label
	label = '$(b)$'

	## input	
	l  = .01
	le = .01

	## input
	R  = .2

	## input
	L = 5.

	## accrete
	Tacc = 1.25


	## vv
	voff = -.4*Tacc - 0.

	## params
	seed = 0
	mpgo=True
	params = dict()
	chainparams = dict()

##################################

	## func
	func0 = xh.mf.dS(L=1.*L)
	func1 = xh.mf.Hay_dS(l=1.*l, R=1.*R, L=1.*L)
	func2 = xh.mf.Hay_dS_cutoff(l=1.*l, R=1.*R, L=1.*L)
	func3 = xh.mf.dS(L=1.*L)
	funclist = [func0, func1, func2, func3]

	## need to match 1 and 2 at infinty
	c12 = func2.F(10.*L) - func1.F(10.*L)

	## regions
	deflines = False
	reg0 = xh.reg.MAXreg(func0, boundary=deflines, rlines=deflines)
	reg1 = xh.reg.EFreg( func1, boundary=deflines, rlines=deflines)
	reg2 = xh.reg.MAXreg(func2, rparams=dict(c=1.*c12), boundary=deflines, rlines=deflines)
	reg3 = xh.reg.MAXreg(func3, boundary=deflines, rlines=deflines)
	reglist = [reg0, reg1, reg2, reg3]

	## modify Hay-AdS region to add extra blocks
	## top
	b  = reg1.blocks[3]
	cdlu = 1.*b.bparams['cdlu'] + 1.
	cdlv = 1.*b.bparams['cdlv'] + 1.
	epsu = -1.*b.bparams['epsu']
	epsv = -1.*b.bparams['epsv']
	bparams = dict(cdlu=1.*cdlu, cdlv=1.*cdlv, epsu=1.*epsu, epsv=1.*epsv)
	reg1.blocks += [xh.block(b.master, b.j, dict(cdlu=1.*cdlu, cdlv=1.*cdlv, epsu=1.*epsu, epsv=1.*epsv))]
	## right
	b  = reg1.blocks[2]
	cdlu = 1.*b.bparams['cdlu'] - 1.
	cdlv = 1.*b.bparams['cdlv'] + 1.
	epsu = -1.*b.bparams['epsu']
	epsv = -1.*b.bparams['epsv']
	bparams = dict(cdlu=1.*cdlu, cdlv=1.*cdlv, epsu=1.*epsu, epsv=1.*epsv)
	reg1.blocks += [xh.block(b.master, b.j, dict(cdlu=1.*cdlu, cdlv=1.*cdlv, epsu=1.*epsu, epsv=1.*epsv))]


	## chain params
	m  = np.array([0., 0.5*R, 0., 0.])
	Rh = np.array([0., func1.rj[2], 0., 0.])
	va = 1.*voff
	vb = 1.*va + 1.*Tacc

	#### initial region
	## passive slice
	sliceloc0 = dict(v0=1.*va, r0=3.*L)
	pslice0 = xh.junc.pslice(reg0, ublocks=[1,2], vblocks=[0,1], **sliceloc0)

	#### Hay-AdS region
	## active slice
	aslice1 = xh.junc.aslice(reg1, ublocks=[3,5], vblocks=[0,1,2,3], U0=pslice0.U_of_r_at_v0, V0=pslice0.V_of_r_at_u0, r_refs=[pslice0.reg.metfunc.r_ref], **sliceloc0)
	## modify transformations
	reg1.U_of_udl = aslice1.U_of_udl_at_v0
	reg1.V_of_vdl = aslice1.V_of_vdl_at_u0
	## passive slice
	sliceloc1 = dict(v0=1.*vb, r0=Rh[1]+le)
	pslice1 = xh.junc.pslice(reg1, ublocks=[2,4], vblocks=[0,1,2,3], **sliceloc1)

	#### Cutoff region
	## modify transformations
	reg2.U_of_udl = reg1.U_of_udl
	reg2.V_of_vdl = reg1.V_of_vdl

	#### Final region
	## active slice
	aslice3 = xh.junc.aslice(reg3, ublocks=[0,3], vblocks=[0,1], U0=pslice1.U_of_r_at_v0, V0=pslice1.V_of_r_at_u0, r_refs=[pslice1.reg.metfunc.r_ref], **sliceloc1)
	ub = 1.*aslice3.u0
	## modify transformations
	reg3.U_of_udl = aslice3.U_of_udl_at_v0
	reg3.V_of_vdl = aslice3.V_of_vdl_at_u0


	#### mask
	## initial region
	reg0.blocks = reg0.blocks[0:2]
	for b in reg0.blocks:
		b.uvbounds.update(dict(vmax=1.*va))
	# Cutoff region
	reg2.blocks = reg2.blocks[2:3]
	reg2.blocks[0].bparams.update(reg1.blocks[-1].bparams)
	reg2.blocks[0].update()
	# Hay-AdS region past slice
	reg1.blocks = reg1.blocks[0:5]
	for b in reg1.blocks[0:4]:
		b.uvbounds.update(dict(vmin=1.*va))
	# Hay-AdS future slice
	newblocks = []
	for i in [0,1]:
		newblocks += [xh.block(reg1.blocks[i].master, reg1.blocks[i].j, reg1.blocks[i].bparams)]
		newblocks[-1].uvbounds.update(reg1.blocks[i].uvbounds)
		newblocks[-1].uvbounds.update(dict(vmax=1.*vb))
	for i in [2]:
		newblocks += [xh.block(reg1.blocks[i].master, reg1.blocks[i].j, reg1.blocks[i].bparams)]
		newblocks[-1].uvbounds.update(reg1.blocks[i].uvbounds)
		newblocks[-1].uvbounds.update(dict(umax=1.*ub, vmax=1.*vb))
	for i in [2]:
		newblocks += [xh.block(reg1.blocks[i].master, reg1.blocks[i].j, reg1.blocks[i].bparams)]
		newblocks[-1].uvbounds.update(reg1.blocks[i].uvbounds)
		newblocks[-1].uvbounds.update(dict(umax=1.*ub, vmin=1.*vb))
	for i in [4]:
		newblocks += [xh.block(reg1.blocks[i].master, reg1.blocks[i].j, reg1.blocks[i].bparams)]
		newblocks[-1].uvbounds.update(reg1.blocks[i].uvbounds)
		newblocks[-1].uvbounds.update(dict(umax=1.*ub))
	for i in [3]:
		newblocks += [xh.block(reg1.blocks[i].master, reg1.blocks[i].j, reg1.blocks[i].bparams)]
		newblocks[-1].uvbounds.update(reg1.blocks[i].uvbounds)
		newblocks[-1].uvbounds.update(dict())	
	## reset Hay-AdS blocks to masked ones
	reg1.blocks = newblocks
	## final region
	reg3.blocks = [reg3.blocks[0],reg3.blocks[3]]
	reg3.blocks[0].uvbounds.update(dict(vmin=1.*vb, umin=1.*ub))
	reg3.blocks[1].uvbounds.update(dict(umin=1.*ub))

	## chainparams
	m = 1.*m
	Rh = 1.*Rh
	fs_v0 = np.array([1.*va, 1.*vb, 0., 0.])
	fs_u0 = np.array([0., 1.*ub, 0., 0.])
	fs_r0 = 1.*Rh+le
	chainparams.update(dict(m=1.*m, Rh=1.*Rh, fs_r0=1.*fs_r0, fs_u0=1.*fs_u0, fs_v0=1.*fs_v0))

	print(chainparams)

######################################

	## draw
	if draw:
		print("plot")
		pp = dict(l=1.*l, R=1.*R)
		xh.evap.draw_dS.drawreg(reglist, chainparams, fparams=pp)

	## label
	if True:
		plt.annotate(s=label, xy=(.95,.97), xycoords='axes fraction', ha='right', va='top', size=8)


	## auto param label
	if True:
		#plabel = [r"", r"$\tau_{acc}=%3s$"%(Tacc), r"$\tau_{ev}=%3s$"%(Tevap)]
		plabel = [r"$l_{ev}=%3s$"%(le), r"", r"$l=%3s$"%(l), r"$2M=%3s$"%(R), r"$L=%3s$"%(L)]
		plabel = "\n".join(plabel)
		plt.annotate(s=plabel, xy=(.95,.03), xycoords='axes fraction', ha='right', va='bottom', size=8)

	## manual param label
	if False:
		plabel = [r"$\tau_{ev}=%s$"%(Tevap), r"$\tau_{acc}=%s$"%(Tacc)]
		plabel += [r"", r"$M=10^{-3}$", r"$l_{ev}=l=10^{-4}$"]
		plabel = "\n".join(plabel)
		plt.annotate(s=plabel, xy=(.95,.03), xycoords='axes fraction', ha='right', va='bottom', size=8)

	## r=infty label
	if True:
		plt.annotate(s=r"$r=\infty$", xy=(0., 1.), ha='center', va='center', size=7)
		plt.annotate(s=r"$r=\infty$", xy=(0.,-1.05), ha='center', va='center', size=7)
		plt.annotate(s=r"$r=0$", xy=(-1.1, 0.), ha='center', va='center', size=7, rotation=90)
		plt.annotate(s=r"$r=0$", xy=(1.02, 0.), ha='center', va='center', size=7, rotation=-90)

	## save
	if True:
		xh.evap.evapsave(path=path, params=params, chainparams=chainparams, seed=seed, sfp=sfp, temp_only=temp_only, massplot=mpgo)


	## show
	if False:
		plt.show()


	## return
	return reglist, chainparams




if __name__=="__main__":
	demo()






"""
Tests for shell_junction modules.
"""


import numpy as np

from xhorizon.shell_junction.active_slice_class import active_slice
from xhorizon.shell_junction.passive_slice_class import passive_slice
from xhorizon.shell_junction.helpers import *




"""
Tests. Run when __name__=='__main__'.
"""


def test1():
	"""
	Test uvdl_of_r_at_uv0.
	"""
	##
	print("\nTEST 1\n")
	## params
	func = xh.mf.hayward()
	ublocks, vblocks = [2,5,3], [0,1,2]
	u0, v0 = 3.*np.random.randn(2)
	## go
	reg = xh.reg.MAXreg(func)
	r = reg.metfunc.r_ref
	#r = r[r<10.]
	r, uvdl_u0, uvdl_v0 = uvdl_of_r_at_uv0(r, reg, ublocks=ublocks, vblocks=vblocks, v0=v0, u0=u0)
	## print
	print("\n")
	print("u0, v0 = %s, %s"%(u0,v0))
	print("r, uvdl_u0, uvdl_v0 = ...")
	print(r)
	print(uvdl_u0)
	print(uvdl_v0)
	print("\n")
	## plot
	plt.figure(1, figsize=(12,5))
	## left subfig
	plt.subplot(121)
	plt.gca().set_aspect('equal')
	## plot region
	reg.rplot()
	## fill blocks
	cols = ['b','y']
	for b in bcon(reg,ublocks):
		sty = dict(fc=cols[0], alpha=0.3)
		b.fill(sty=sty)
	for b in bcon(reg,vblocks):
		sty = dict(fc=cols[1], alpha=0.3)
		b.fill(sty=sty)
	## plot slices
	i = 0
	cmaps = [plt.cm.Blues, plt.cm.Reds]
	for uvdl in [uvdl_u0, uvdl_v0]:
		x, y = uvdl[1] - uvdl[0], uvdl[1] + uvdl[0]
		plt.scatter(x, y, marker='x', s=5, c=-r, cmap=cmaps[i%len(cmaps)], zorder=5000)
		i+=1
	## right subfig
	plt.subplot(122)
	plt.xlabel('r')
	plt.grid()
	## plot slices
	i = 0
	cmaps = [plt.cm.Blues, plt.cm.Reds]
	for uvdl in [uvdl_u0, uvdl_v0]:
		plt.scatter(r, uvdl[(i+0)%2], marker='x', s=1, c=-r, cmap=cmaps[i%2], zorder=5000)
		plt.scatter(r, uvdl[(i+1)%2], marker='x', s=10, c=-r, cmap=cmaps[i%2], zorder=5000)
		i+=1
	##
	plt.show()
	##
	print("\nEND TEST 1\n")



def test2():
	"""
	Test set_ruv0.
	"""
	##
	print("\nTEST 2\n")
	##
	reg = xh.reg.MAXreg(xh.mf.minkowski())
	r0, u0, v0 = set_ruv0(reg, r0=5, u0=8)
	print('r0=%s, u0=%s, v0=%s'%(r0,u0,v0))
	##
	print("\nEND TEST 2\n")





def test3():
	"""
	Test active_slice plotting uvdl ref values.
	"""
	##
	print("\nTEST 3\n")
	## define func and region
	func = xh.mf.schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=False)
	## set input params and evaluate
	r0, u0, v0 = 1.1, np.nan, -2.
	ublocks, vblocks = [1,2], [0,1]
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0)
	##
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	## plot
	plt.figure(1, figsize=(12,5))
	## left subfig
	plt.subplot(121)
	plt.gca().set_aspect('equal')
	## plot region
	reg.rplot()
	## fill blocks
	cols = ['b','y']
	for b in bcon(reg,ublocks):
		sty = dict(fc=cols[0], alpha=0.3)
		b.fill(sty=sty)
	for b in bcon(reg,vblocks):
		sty = dict(fc=cols[1], alpha=0.3)
		b.fill(sty=sty)
	## plot ref array slices
	if True:
		cvals = -np.arange(len(slc.r))
		cmaps = [plt.cm.Blues, plt.cm.Reds]
		plt.scatter(slc.uvdl_u0[1]-slc.uvdl_u0[0],slc.uvdl_u0[1]+slc.uvdl_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
		plt.scatter(slc.uvdl_v0[1]-slc.uvdl_v0[0],slc.uvdl_v0[1]+slc.uvdl_v0[0],marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
	## right subfig
	plt.subplot(122)
	plt.xlabel('r')
	plt.grid()
	## plot slices
	if True:
		i = 0
		cvals = -np.arange(len(slc.r))
		cmaps = [plt.cm.Blues, plt.cm.Reds]
		for uvdl in [slc.uvdl_u0, slc.uvdl_v0]:
			plt.scatter(slc.r, uvdl[(i+0)%2], marker='x', s=1 , c=cvals, cmap=cmaps[i%2], zorder=5000)
			plt.scatter(slc.r, uvdl[(i+1)%2], marker='x', s=10, c=cvals, cmap=cmaps[i%2], zorder=5000)
			i+=1
	##
	plt.show()
	print("\nEND TEST 3\n")





def test4():
	"""
	Test active_slice UV ref values.
	"""
	##
	print("\nTEST 4\n")
	## define func and region
	func = xh.mf.schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=False)
	## set input params and evaluate
	U0 = lambda r: - np.pi**(-1) * np.arctan(r)
	V0 = lambda r:   np.pi**(-1) * np.arctan(r)
	r0, u0, v0 = 1.5, np.nan, -2.
	ublocks, vblocks = [1,2], [0,1]
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, U0=U0, V0=V0)
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	##
	## plot
	if True:
		plt.figure(1, figsize=(12,5))
		## left subfig
		plt.subplot(121)
		plt.gca().set_aspect('equal')
		## plot region
		reg.rplot()
		## fill blocks
		cols = ['b','y']
		for b in bcon(reg,ublocks):
			sty = dict(fc=cols[0], alpha=0.3)
			b.fill(sty=sty)
		for b in bcon(reg,vblocks):
			sty = dict(fc=cols[1], alpha=0.3)
			b.fill(sty=sty)
		## plot ref array slices
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			## u0,V0 slice by replacing vdl_u0=uvdl_u0[1] by V_u0
			plt.scatter(slc.V_u0-slc.uvdl_u0[0],slc.V_u0+slc.uvdl_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
			## v0,U0 slice by replacing udl_v0=uvdl_v0[0] by U_v0
			plt.scatter(slc.uvdl_v0[1]-slc.U_v0,slc.uvdl_v0[1]+slc.U_v0,marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
		## right subfig
		plt.subplot(122)
		plt.xlabel('r')
		plt.grid()
		## plot slices
		if False:
			i = 0
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			for uvdl in [slc.uvdl_u0, slc.uvdl_v0]:
				plt.scatter(slc.r, uvdl[(i+0)%2], marker='x', s=1 , c=cvals, cmap=cmaps[i%2], zorder=5000)
				plt.scatter(slc.r, uvdl[(i+1)%2], marker='x', s=10, c=cvals, cmap=cmaps[i%2], zorder=5000)
				i+=1
		##
		plt.show()
	print("\nEND TEST 4\n")



def test5():
	"""
	Test active_slice.
	"""
	## import
	## start
	print("\nTEST 6\n")
	## define func and region
	func = xh.mf.schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=True)
	## set input params and evaluate
	U0 = lambda r: - np.pi**(-1) * np.arctan(r)
	V0 = lambda r:   np.pi**(-1) * np.arctan(r)
	r0, u0, v0 = 2.1, np.nan, 2.
	ublocks, vblocks = [1,2], [0,1]
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, U0=U0, V0=V0)
	## play with slice object
	if False:
		plt.figure(1)
		sq = 1.5
		plt.xlim(-sq,sq)
		plt.ylim(-sq,sq)
		plt.gca().set_aspect('equal')
		plt.grid()
		plt.plot(slc.uvdl_v0[0], slc.UV_v0[0], 'rx')
		plt.plot(slc.uvdl_u0[1], slc.UV_u0[1], 'bx')
		plt.show()
	## done
	print("\nEND TEST 5\n")




def test6():
	"""
	Test active_slice.
	Schwarzschild with slice outside horizon.
	"""
	##
	print("\nTEST 6\n")
	## define func and region
	func = xh.mf.schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=True)
	## set input params
	U0 = lambda r: - np.pi**(-1) * np.arctan(r)
	V0 = lambda r:   np.pi**(-1) * np.arctan(r)
	## r>1
	r0, u0, v0 = 2.1, 4., np.nan
	ublocks, vblocks = [1,2], [0,1]
	## evaluate slice
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, U0=U0, V0=V0, mu=0.)
	## update region UV transformations
	if True:
		reg.U_of_udl = slc.U_of_udl_at_v0
		reg.V_of_vdl = slc.V_of_vdl_at_u0
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	##
	## plot
	if True:
		plt.figure(1, figsize=(12,5))
		## left subfig
		plt.subplot(121)
		plt.gca().set_aspect('equal')
		## plot region
		reg.rplot()
		## fill blocks
		cols = ['b','y']
		for b in bcon(reg,ublocks):
			sty = dict(fc=cols[0], alpha=0.3)
			b.fill(sty=sty)
		for b in bcon(reg,vblocks):
			sty = dict(fc=cols[1], alpha=0.3)
			b.fill(sty=sty)
		## plot ref array slices
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			## plot region in UV coords
			plt.scatter(slc.UV_u0[1]-slc.UV_u0[0],slc.UV_u0[1]+slc.UV_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
			plt.scatter(slc.UV_v0[1]-slc.UV_v0[0],slc.UV_v0[1]+slc.UV_v0[0],marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
		## right subfig
		plt.subplot(122)
		plt.xlabel('uvdl')
		plt.ylabel('UV')
		plt.grid()
		## plot interpolated U(udl) and V(vdl)
		if True:
			xdl = 1.1*np.linspace(-1,1,5001)
			plt.scatter(xdl, slc.V_of_vdl_at_u0(xdl), marker='x', s=5, linestyle='-', c='b', zorder=5000)
			plt.scatter(xdl, slc.U_of_udl_at_v0(xdl), marker='x', s=5, linestyle='-', c='r', zorder=5000)
			
		## plot reference values U_v0(udl_v0) and V_v0(vdl_v0)
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			plt.scatter(slc.uvdl_u0[1], slc.UV_u0[1], marker='x', s=10 , c=cvals, cmap=cmaps[0], zorder=5100)
			plt.scatter(slc.uvdl_v0[0], slc.UV_v0[0], marker='x', s=10 , c=cvals, cmap=cmaps[1], zorder=5100)
		##
		plt.show()
	print("\nEND TEST 6\n")




def test7():
	"""
	Test active_slice.
	Schwarzschild with slice inside horizon.
	"""
	##
	print("\nTEST 7\n")
	## define func and region
	func = xh.mf.schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=True)
	## set input params
	U0 = lambda r: - np.pi**(-1) * np.arctan(r)
	V0 = lambda r: - np.pi**(-1) * np.arctan(r)
	## r>1
	r0, u0, v0 = 0.5, 4., np.nan
	ublocks, vblocks = [0,3], [0,1]
	## evaluate slice
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, U0=U0, V0=V0, mu=0.)
	## update region UV transformations
	if True:
		reg.U_of_udl = slc.U_of_udl_at_v0
		reg.V_of_vdl = slc.V_of_vdl_at_u0
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	##
	## plot
	if True:
		plt.figure(1, figsize=(12,5))
		## left subfig
		plt.subplot(121)
		plt.xlim(-1,1)
		plt.ylim(-1,1)
		plt.gca().set_aspect('equal')
		plt.grid()
		## plot region
		reg.rplot()
		## fill blocks
		cols = ['b','y']
		for b in bcon(reg,ublocks):
			sty = dict(fc=cols[0], alpha=0.3)
			b.fill(sty=sty)
		for b in bcon(reg,vblocks):
			sty = dict(fc=cols[1], alpha=0.3)
			b.fill(sty=sty)
		## plot ref array slices
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			## plot region in UV coords
			plt.scatter(slc.UV_u0[1]-slc.UV_u0[0],slc.UV_u0[1]+slc.UV_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
			plt.scatter(slc.UV_v0[1]-slc.UV_v0[0],slc.UV_v0[1]+slc.UV_v0[0],marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
		## right subfig
		plt.subplot(122)
		plt.xlabel('uvdl')
		plt.ylabel('UV')
		plt.grid()
		## plot interpolated U(udl) and V(vdl)
		if True:
			xdl = 1.1*np.linspace(-1,1,5001)
			plt.scatter(xdl, slc.V_of_vdl_at_u0(xdl), marker='x', s=5, linestyle='-', c='b', zorder=5000)
			plt.scatter(xdl, slc.U_of_udl_at_v0(xdl), marker='x', s=5, linestyle='-', c='r', zorder=5000)
			
		## plot reference values U_v0(udl_v0) and V_v0(vdl_v0)
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			plt.scatter(slc.uvdl_u0[1], slc.UV_u0[1], marker='x', s=10 , c=cvals, cmap=cmaps[0], zorder=5100)
			plt.scatter(slc.uvdl_v0[0], slc.UV_v0[0], marker='x', s=10 , c=cvals, cmap=cmaps[1], zorder=5100)
		##
		plt.show()
	print("\nEND TEST 7\n")



def test8():
	"""
	Test active_slice.
	Hayward with slice outside horizons.
	"""
	##
	print("\nTEST 8\n")
	## define func and region
	func = xh.mf.hayward()
	reg = xh.reg.MAXreg(func,rlines=True)
	## set input params
	U0 = lambda r: - np.pi**(-1) * np.arctan(r)
	V0 = lambda r:   np.pi**(-1) * np.arctan(r)
	## r>1
	r0, u0, v0 = 2.5, 1., np.nan
	ublocks, vblocks = [2,5,3], [0,1,2]
	## evaluate slice
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, U0=U0, V0=V0, mu=0.1)
	## update region UV transformations
	if True:
		reg.U_of_udl = slc.U_of_udl_at_v0
		reg.V_of_vdl = slc.V_of_vdl_at_u0
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	##
	## plot
	if True:
		plt.figure(1, figsize=(12,5))
		## left subfig
		plt.subplot(121)
		plt.xlim(-2,2)
		plt.ylim(-2,2)
		plt.gca().set_aspect('equal')
		plt.grid()
		## plot region
		reg.rplot()
		## fill blocks
		cols = ['b','y']
		for b in bcon(reg,ublocks):
			sty = dict(fc=cols[0], alpha=0.3)
			b.fill(sty=sty)
		for b in bcon(reg,vblocks):
			sty = dict(fc=cols[1], alpha=0.3)
			b.fill(sty=sty)
		## plot ref array slices
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			## plot region in UV coords
			plt.scatter(slc.UV_u0[1]-slc.UV_u0[0],slc.UV_u0[1]+slc.UV_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
			plt.scatter(slc.UV_v0[1]-slc.UV_v0[0],slc.UV_v0[1]+slc.UV_v0[0],marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
		## right subfig
		plt.subplot(122)
		plt.xlabel('uvdl')
		plt.ylabel('UV')
		plt.grid()
		## plot interpolated U(udl) and V(vdl)
		if True:
			xdl = 2.5*np.linspace(-1,1,5001)
			plt.scatter(xdl, slc.V_of_vdl_at_u0(xdl), marker='x', s=5, linestyle='-', c='b', zorder=5000)
			plt.scatter(xdl, slc.U_of_udl_at_v0(xdl), marker='x', s=5, linestyle='-', c='r', zorder=5000)
			
		## plot reference values U_v0(udl_v0) and V_v0(vdl_v0)
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			plt.scatter(slc.uvdl_u0[1], slc.UV_u0[1], marker='x', s=10 , c=cvals, cmap=cmaps[0], zorder=5100)
			plt.scatter(slc.uvdl_v0[0], slc.UV_v0[0], marker='x', s=10 , c=cvals, cmap=cmaps[1], zorder=5100)
		##
		plt.show()
	print("\nEND TEST 8\n")



def test9():
	"""
	Test active_slice.
	Check shell functionality when U0 or V0 is None.
	Seems to be working perfectly.
	Both None leaves UV=uvdl. One or the other does the shell transformation. Both does corner.
	"""
	##
	print("\nTEST 9\n")
	## define func and region
	func = xh.mf.schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=True)
	## set input params
	U0 = lambda r: - np.pi**(-1) * np.arctan(r)
	V0 = lambda r:   np.pi**(-1) * np.arctan(r)
	## r>1
	r0, u0, v0 = 2.053, np.nan, -1.5
	ublocks, vblocks = [1,2], [0,1]
	## evaluate slice
	slc = active_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, U0=None, V0=V0, mu=0.)
	## update region UV transformations
	if True:
		reg.U_of_udl = slc.U_of_udl_at_v0
		reg.V_of_vdl = slc.V_of_vdl_at_u0
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	##
	## plot
	if True:
		plt.figure(1, figsize=(12,5))
		## left subfig
		plt.subplot(121)
		plt.gca().set_aspect('equal')
		## plot region
		reg.rplot()
		## fill blocks
		cols = ['b','y']
		for b in bcon(reg,ublocks):
			sty = dict(fc=cols[0], alpha=0.3)
			b.fill(sty=sty)
		for b in bcon(reg,vblocks):
			sty = dict(fc=cols[1], alpha=0.3)
			b.fill(sty=sty)
		## plot ref array slices
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			## plot region in UV coords
			plt.scatter(slc.UV_u0[1]-slc.UV_u0[0],slc.UV_u0[1]+slc.UV_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
			plt.scatter(slc.UV_v0[1]-slc.UV_v0[0],slc.UV_v0[1]+slc.UV_v0[0],marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
		## right subfig
		plt.subplot(122)
		plt.xlabel('uvdl')
		plt.ylabel('UV')
		plt.grid()
		## plot interpolated U(udl) and V(vdl)
		if True:
			xdl = 1.1*np.linspace(-1,1,5001)
			plt.scatter(xdl, slc.V_of_vdl_at_u0(xdl), marker='x', s=5, linestyle='-', c='b', zorder=5000)
			plt.scatter(xdl, slc.U_of_udl_at_v0(xdl), marker='x', s=5, linestyle='-', c='r', zorder=5000)
			
		## plot reference values U_v0(udl_v0) and V_v0(vdl_v0)
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			plt.scatter(slc.uvdl_u0[1], slc.UV_u0[1], marker='x', s=10 , c=cvals, cmap=cmaps[0], zorder=5100)
			plt.scatter(slc.uvdl_v0[0], slc.UV_v0[0], marker='x', s=10 , c=cvals, cmap=cmaps[1], zorder=5100)
		##
		plt.show()
	print("\nEND TEST 9\n")



def test10():
	"""
	Test passive_slice.
	Schwarz, outside horizon: working fine.
	Schwarz, inside horizon: working fine.
	Hayward, outside horizon: working fine.
	"""
	##
	print("\nTEST 10\n")
	## define func and region
	func = xh.mf.hayward()   ##schwarzschild()
	reg = xh.reg.MAXreg(func,rlines=True)
	## r>1
	r0, u0, v0 = 1.3, np.nan, 0.5
	ublocks, vblocks =  [2,5,3], [0,1,2]  ##[0,3], [0,1]
	##
	if True:
		reg.U_of_udl = lambda udl: np.arctan(udl)
		reg.V_of_vdl = lambda vdl: np.arctan(vdl)
	## evaluate slice
	slc = passive_slice(reg, ublocks=ublocks, vblocks=vblocks, r0=r0, u0=u0, v0=v0, mu=0.)
	## add ref lines to reg
	sty = sty=dict(c='lime', lw=1, alpha=0.9, zorder=6000)
	for b in slc.ublocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(u0):
			b.add_curves_uv(xh.cm.uvlines([u0], uv='u', sty=sty))
	for b in slc.vblocks:
		if np.isfinite(r0):
			b.add_curves_tr(xh.cm.rlines([r0], sty=sty))
		if np.isfinite(v0):
			b.add_curves_uv(xh.cm.uvlines([v0], uv='v', sty=sty))
	##
	## plot
	if True:
		plt.figure(1, figsize=(12,5))
		## left subfig
		plt.subplot(121)
		plt.gca().set_aspect('equal')
		## plot region
		reg.rplot()
		## fill blocks
		cols = ['b','y']
		for b in bcon(reg,ublocks):
			sty = dict(fc=cols[0], alpha=0.3)
			b.fill(sty=sty)
		for b in bcon(reg,vblocks):
			sty = dict(fc=cols[1], alpha=0.3)
			b.fill(sty=sty)
		## plot ref array slices
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			## plot region in UV coords
			plt.scatter(slc.UV_u0[1]-slc.UV_u0[0],slc.UV_u0[1]+slc.UV_u0[0],marker='x',s=10,c=cvals,cmap=cmaps[0],zorder=5000)
			plt.scatter(slc.UV_v0[1]-slc.UV_v0[0],slc.UV_v0[1]+slc.UV_v0[0],marker='x',s=10,c=cvals,cmap=cmaps[1],zorder=5000)
		## right subfig
		plt.subplot(122)
		plt.xlabel('r')
		plt.ylabel('UV')
		#plt.xlim(0,2)
		plt.grid()
		## plot interpolated U(r) and V(r)
		if True:
			rr = 5.*np.linspace(0,1,5001)
			plt.scatter(rr, slc.V_of_r_at_u0(rr), marker='x', s=5, linestyle='-', c='b', zorder=5000)
			plt.scatter(rr, slc.U_of_r_at_v0(rr), marker='x', s=5, linestyle='-', c='r', zorder=5000)
			
		## plot reference values U_v0(r) and V_v0(r)
		if True:
			cvals = -np.arange(len(slc.r))
			cmaps = [plt.cm.Blues, plt.cm.Reds]
			plt.scatter(slc.r, slc.UV_u0[1], marker='x', s=10 , c=cvals, cmap=cmaps[0], zorder=5100)
			plt.scatter(slc.r, slc.UV_v0[0], marker='x', s=10 , c=cvals, cmap=cmaps[1], zorder=5100)
		##
		plt.show()
	print("\nEND TEST 10\n")




if __name__=="__main__":

	import matplotlib.pyplot as plt
	import xhorizon as xh

	print("\n\nTESTS\n")
	test9()
	print("\nEND TESTS\n")






from xhorizon.diagram_tools import curvemakers
from xhorizon.diagram_tools import coord_transf
from xhorizon.diagram_tools import block_masks
from xhorizon.diagram_tools import block_fill
from xhorizon.diagram_tools import curve_class
from xhorizon.diagram_tools import diagram_classes
from xhorizon.diagram_tools import plotstyle
from xhorizon.diagram_tools import region_factory




"""
Functions used to interpolate and extrapolate coord transformations from reference arrays.
"""

import numpy as np




def interp_with_smooth_extrap(x, x_ref, y_ref, mu=0.):
	"""
	Interpolate with smooth extrapolation from the reference endpoints,
	approaching a line of unit slope as mu*x goes to infinity.

	A wrapper for numpy.interp allowing:
		a. custom handling of x values outside the x_ref domain,
		b. scrubbing nan values and sorting ref arrays.
	
	The smooth continuation and decay to unit slope line uses the function 
		f1(x) = x + (mu/b)*(s0-1.)*(1.-np.exp(-x*b/mu)).

	We are specifically working with monotonic functions, which allows us to know whether to
	approach a slope of +1 or -1. The algorithm is basically tailored to monotonic increasing
	functions. Monotonic decreasing functions are handled by vertically flipping the reference
	y_ref array, then proceeding with the monotonic increasing routine. The final result y is
	then flipped back at the end.

	Output y(x) is continuous and once differentiable.

	Parameter mu multiplies the decay length for exponentally approaching unit slope line.
	When mu=0 the extrapolation is not differentiable, the line of extrapolation is just y=x.
	As mu goes to infinity, the line of extrapolation becomes linear, and this function 
	approaches interp_with_linear_extrap.
	"""
	## init
	x, x_ref, y_ref = 1.*x, 1.*x_ref, 1.*y_ref
	mu = np.abs(1.*float(mu))
	y = np.nan*x
	## scrub refs
	mask = np.logical_and.reduce( [np.isfinite(x_ref), np.isfinite(y_ref)] )
	x_ref, y_ref = x_ref[mask], y_ref[mask]
	## sort refs
	idx = np.argsort(x_ref)
	x_ref, y_ref = x_ref[idx], y_ref[idx]
	## flip if decreasing
	flip = False
	if y_ref[-1]<y_ref[0]:
		flip = True
		y_ref = -y_ref
	## get first and last ref points
	x0, y0 = x_ref[0],  y_ref[0]
	x1, y1 = x_ref[-1], y_ref[-1]
	## get slopes at first and last ref points
	dy0 = get_dy0(x_ref, y_ref)
	dy1 = get_dy1(x_ref, y_ref)
	## interpolate
	mask = np.logical_and.reduce( [x0<x, x<x1] )
	y[mask] = np.interp(x[mask], x_ref, y_ref)
	## extrapolate
	## left
	mask = x<=x0
	y[mask] = y0 + f1((x[mask]-x0), s0=dy0, mu=-mu, b=0.1+np.abs(np.log(dy0)))
	## right
	mask = x>=x1
	y[mask] = y1 + f1((x[mask]-x1), s0=dy1, mu= mu, b=0.1+np.abs(np.log(dy1)))
	## unflip if yref was flipped
	if flip==True:
		y = -y
	## return
	return y



def f1(x, s0=1., mu=0., b=1.):
	"""
	Monotonic function f(x) passing through origin with slope s0, 
	such that slope decays to 1 as mu*x goes to infinity.
	Inputs:
		x =  An array of x values to evaluate.
		s0 = Slope at the origin, used for differentiable matching.
		mu = Exponential decay length is mu/b. Typically use mu as overall user controlled multiplier.
		b  = Exponential decay length is mu/b. Typically use b to set an algorithmic length scale control.
			e.g. setting b = |ln(s0)| causes a stronger exponential decay when s0>>1 or s0<<1.
		Parameters mu and b are separate only for convenience and implementation.
	Returns:
		y = Array of values f(x).
	"""
	if np.any(np.logical_not(np.isfinite(np.array([b,s0,1./b])))):
		print('mu, s0, b = %s, %s, %s'%(mu, s0, b))

	## init
	x  = 1.*x
	mu = 1.*float(mu)
	b = 1.*float(b)
	s0 = 1.*float(s0)
	## case mu=0
	if mu==0.:
		y = 1.*x
	## go
	else:
		y = x + (mu/b)*(s0-1.)*(1.-np.exp(-x*b/mu))
	## return
	return y


def get_dy0(x,y):
	"""
	Find initial slope by iterating until finite.
	Avoids issues with repeated initial point.
	"""
	## loop until finite
	i, dy = 0, np.nan
	while (not np.isfinite(dy)) or dy==0.:
		i += 1
		dy  = (y[i] - y[0]) / (x[i] - x[0])
	## return
	return dy

def get_dy1(x,y):
	"""
	Find final slope by iterating until finite.
	Avoids issues with repeated final point.
	"""
	## loop until finite
	i, dy = 1, np.nan
	while (not np.isfinite(dy)) or dy==0.:
		i += 1
		dy  = (y[-1] - y[-i]) / (x[-1] - x[-i])
	## return
	return dy


















"""
Tests. Run if __name__=="__main__".
"""


def test2():
	"""
	Test interp_with_smooth_extrap.
	"""
	##
	print("\nTEST 2\n")
	## params
	A = 1e1
	mux = [.1,1.,10.]
	## ref values
	x_ref = np.linspace(-.8,1.2,1001)
	y_ref = A * x_ref**(3)
	## go
	## plot ref
	plt.plot(x_ref,y_ref,'rx', zorder=5)
	for i in range(len(mux)):
	## interp values
		mu = mux[i]
		x = np.linspace(-5,5,5001)
		y = interp_with_smooth_extrap(x,x_ref,y_ref,mu=mu)
		## plot interp
		c = 1.- float(i) / float(len(mux)) 
		plt.plot(x,y, 'b-', lw=2, alpha=c, zorder=100-i)
	##
	plt.gca().set_aspect('equal')
	sq = 5.
	ymult = 5.
	plt.xlim(-sq,sq)
	plt.ylim(-ymult*sq,ymult*sq)
	plt.grid()
	plt.show()
	##
	print("\nEND TEST 2\n")



def test3():
	"""
	Test interp_with_smooth_extrap.
	Decreasing functions now work.
	"""
	##
	print("\nTEST 3\n")
	## params
	A = -10.
	mu = 1.
	## ref values
	x_ref = np.linspace(-.95,1.05,1001)
	y_ref = A * x_ref**(3)
	## interp values
	x = np.linspace(-5,5,5001)
	y = interp_with_smooth_extrap(x, x_ref, y_ref, mu=mu)
	## plot ref
	plt.plot(x_ref,y_ref,'rx', zorder=5)
	## plot interp
	plt.plot(x,y, 'b-', lw=2, zorder=100)
	##
	plt.gca().set_aspect('equal')
	sq = 5.
	ymult = 2.
	plt.xlim(-sq,sq)
	#plt.ylim(-ymult*sq,ymult*sq)
	plt.grid()
	plt.show()
	##
	print("\nEND TEST 3\n")





if __name__=="__main__":

	import matplotlib.pyplot as plt

	print("\n\nTESTS\n")
	#test2()
	test3()
	print("\nEND TESTS\n")


